#!/usr/bin/python3

#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, Gio, GLib
import gettext, os, pwd, sys, urllib.parse, locale

from revelation import config, data, datahandler, dialog, entry, io, ui, util

_ = gettext.gettext

class Revelation(ui.App):
    "The Revelation application"

    def __init__(self):
        sys.excepthook = self.__cb_exception
        os.umask(0o077)

        gettext.bindtextdomain(config.PACKAGE, config.DIR_LOCALE)
        gettext.textdomain(config.PACKAGE)

        # Gtk.Builder uses the C lib's locale API, accessible with the locale module
        locale.bindtextdomain(config.PACKAGE, config.DIR_LOCALE)
        locale.bind_textdomain_codeset(config.PACKAGE, "UTF-8")

        ui.App.__init__(self, config.APPNAME)
        self.window = None

        resource_path = os.path.join(config.DIR_UI, 'revelation.gresource')
        resource = Gio.Resource.load(resource_path)
        resource._register()

    def do_startup(self):
        Gtk.Application.do_startup(self)
        if not self.window:
            self.window = ui.AppWindow(application=self, title="Revelation")
            self.main_vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL,5)
            self.window.add(self.main_vbox)
        self.add_window(self.window)

    def do_activate(self):

        self.builder = Gtk.Builder()
        self.builder.add_from_resource('/info/olasagasti/revelation/ui/menubar.ui')
        self.menubar = self.builder.get_object("menubar")

        self.popupbuilder = Gtk.Builder()
        self.popupbuilder.add_from_resource('/info/olasagasti/revelation/ui/popup-tree.ui')
        self.popupmenu = self.popupbuilder.get_object("popup-tree")

        self.window.connect("delete-event", self.__cb_quit)

        try:
            self.__init_config()
            self.__init_actions()
            self.__init_facilities()
            self.__init_ui()
            self.__init_states()
            self.__init_dbus()

        except IOError:
            dialog.Error(self.window, _('Missing data files'), _('Some of Revelations system files could not be found, please reinstall Revelation.')).run()
            sys.exit(1)

        except config.ConfigError:
            dialog.Error(self.window, _('Missing configuration data'), _('Revelation could not find its configuration data, please reinstall Revelation.')).run()
            sys.exit(1)

        except ui.DataError:
            dialog.Error(self.window, _('Invalid data files'), _('Some of Revelations system files contain invalid data, please reinstall Revelation.')).run()
            sys.exit(1)


        if len(sys.argv) > 1:
            file = sys.argv[1]

        elif self.config.get_boolean("file-autoload") == True:
            file = self.config.get_string("file-autoload-file")

        else:
            file = ""

        if file != "":
            self.file_open(io.file_normpath(urllib.parse.unquote(file)))


    def __init_config(self):
        "Get configuration schema"

        schema_source = Gio.SettingsSchemaSource.get_default()
        rvl_schema = schema_source.lookup('org.revelation', recursive=True)

        if not rvl_schema:
            schema_source = Gio.SettingsSchemaSource.new_from_directory(config.DIR_GSCHEMAS, schema_source, False)
            rvl_schema=schema_source.lookup('org.revelation', recursive=True)

        if not rvl_schema:
            raise config.ConfigError

        rvl_settings = Gio.Settings.new_full(rvl_schema, None, None)
        self.config = rvl_settings


    def __init_actions(self):
        "Sets up actions"

        # set up placeholders
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("placeholder", group)

        action_menu_edit  = Gio.SimpleAction.new("menu-edit",  None)
        action_menu_entry = Gio.SimpleAction.new("menu-entry", None)
        action_menu_file  = Gio.SimpleAction.new("menu-file",  None)
        action_menu_help  = Gio.SimpleAction.new("menu-help",  None)
        action_menu_view  = Gio.SimpleAction.new("menu-view",  None)
        action_popup_tree = Gio.SimpleAction.new("popup-tree", None)
        group.add_action(action_menu_edit)
        group.add_action(action_menu_entry)
        group.add_action(action_menu_file)
        group.add_action(action_menu_help)
        group.add_action(action_menu_view)
        group.add_action(action_popup_tree)

        # set up dynamic actions
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("dynamic", group)

        action = Gio.SimpleAction.new("clip-paste", None)
        action.connect("activate",      self.__cb_clip_paste)
        group.add_action(action)

        action = Gio.SimpleAction.new("entry-goto", None)
        action.connect("activate",      lambda w,k: self.entry_goto(self.tree.get_selected()))
        group.add_action(action)

        action = Gio.SimpleAction.new("redo", None)
        action.connect("activate",      lambda w,k: self.redo())
        group.add_action(action)

        action = Gio.SimpleAction.new("undo", None)
        action.connect("activate",      lambda w,k: self.undo())
        group.add_action(action)

        # set up group for multiple entries
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("entry-multiple", group)

        action = Gio.SimpleAction.new("clip-copy", None)
        action.connect("activate",      self.__cb_clip_copy)
        group.add_action(action)

        action = Gio.SimpleAction.new("clip-chain", None)
        action.connect("activate",      lambda w,k: self.clip_chain(self.entrystore.get_entry(self.tree.get_active())))
        group.add_action(action)

        action = Gio.SimpleAction.new("clip-cut", None)
        action.connect("activate",      self.__cb_clip_cut)
        group.add_action(action)

        action = Gio.SimpleAction.new("entry-remove", None)
        action.connect("activate",      lambda w,k: self.entry_remove(self.tree.get_selected()))
        group.add_action(action)

        # action group for "optional" entries
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("entry-optional", group)

        action = Gio.SimpleAction.new("entry-add", None)
        action.connect("activate",      lambda w,k: self.entry_add(None, self.tree.get_active()))
        group.add_action(action)

        action = Gio.SimpleAction.new("entry-folder", None)
        action.connect("activate",      lambda w,k: self.entry_folder(None, self.tree.get_active()))
        group.add_action(action)

        # action group for single entries
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("entry-single", group)

        action = Gio.SimpleAction.new("entry-edit", None)
        action.connect("activate",      lambda w,k: self.entry_edit(self.tree.get_active()))
        group.add_action(action)

        # action group for existing file
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("file-exists", group)

        action = Gio.SimpleAction.new("file-lock", None)
        action.connect("activate",      lambda w,k: self.file_lock())
        group.add_action(action)

        # action group for searching
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("find", group)

        action = Gio.SimpleAction.new("find-next", None)
        action.connect("activate",      lambda w,k: self.__entry_find(self, self.searchbar.entry.get_text(), self.searchbar.dropdown.get_active_type(), data.SEARCH_NEXT))
        group.add_action(action)

        action = Gio.SimpleAction.new("find-previous", None)
        action.connect("activate",      lambda w,k: self.__entry_find(self, self.searchbar.entry.get_text(), self.searchbar.dropdown.get_active_type(), data.SEARCH_PREVIOUS))
        group.add_action(action)

        # global action group
        group = Gio.SimpleActionGroup()
        self.window.insert_action_group("file", group)

        action = Gio.SimpleAction.new("file-change-password", None)
        action.connect("activate",      lambda w,k: self.file_change_password())
        group.add_action(action)

        action = Gio.SimpleAction.new("file-close", None)
        action.connect("activate",      self.__cb_close)
        group.add_action(action)

        action = Gio.SimpleAction.new("file-export", None)
        action.connect("activate",      lambda w,k: self.file_export())
        group.add_action(action)

        action = Gio.SimpleAction.new("file-import", None)
        action.connect("activate",      lambda w,k: self.file_import())
        group.add_action(action)

        action = Gio.SimpleAction.new("file-new", None)
        action.connect("activate",      lambda w,k: self.file_new())
        group.add_action(action)

        action = Gio.SimpleAction.new("file-open", None)
        action.connect("activate",      lambda w,k: self.file_open())
        group.add_action(action)

        action = Gio.SimpleAction.new("file-save", None)
        action.connect("activate",      lambda w,k: self.file_save(self.datafile.get_file(), self.datafile.get_password()))
        group.add_action(action)

        action = Gio.SimpleAction.new("file-save-as", None)
        action.connect("activate",      lambda w,k: self.file_save(None, None))
        group.add_action(action)

        action = Gio.SimpleAction.new("find", None)
        action.connect("activate",      lambda w,k: self.entry_find())
        group.add_action(action)

        action = Gio.SimpleAction.new("help-about", None)
        action.connect("activate",      lambda w,k: self.about())
        group.add_action(action)

        action = Gio.SimpleAction.new("prefs", None)
        action.connect("activate",      lambda w,k: self.prefs())
        group.add_action(action)

        action = Gio.SimpleAction.new("pwchecker", None)
        action.connect("activate",      lambda w,k: self.pwcheck())
        group.add_action(action)

        action = Gio.SimpleAction.new("pwgenerator", None)
        action.connect("activate",      lambda w,k: self.pwgen())
        group.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate",      self.__cb_quit)
        group.add_action(action)

        action = Gio.SimpleAction.new("select-all", None)
        action.connect("activate",      lambda w,k: self.tree.select_all())
        group.add_action(action)

        action = Gio.SimpleAction.new("select-none", None)
        action.connect("activate",      lambda w,k: self.tree.unselect_all())
        group.add_action(action)

        action_vp = Gio.SimpleAction.new_stateful("view-passwords", None, GLib.Variant.new_boolean(self.config.get_boolean("view-passwords")))
        action_vp.connect("activate", lambda w, k: action_vp.set_state(GLib.Variant.new_boolean(not action_vp.get_state())))
        action_vp.connect("activate", lambda w, k: self.config.set_boolean("view-passwords", action_vp.get_state()))
        self.config.connect("changed::view-passwords", lambda w, k: action_vp.set_state(GLib.Variant.new_boolean(w.get_boolean(k))))
        group.add_action(action_vp)

        action_vs = Gio.SimpleAction.new_stateful("view-searchbar", None, GLib.Variant.new_boolean(True))
        action_vs.connect("activate", lambda w, k: action_vs.set_state(GLib.Variant.new_boolean(not action_vs.get_state())))
        action_vs.connect("activate", lambda w, k: self.config.set_boolean("view-searchbar", action_vs.get_state()))
        action_vs.connect("activate", lambda w, k: self.searchbar.set_visible(GLib.Variant.new_boolean(action_vs.get_state())))
        self.config.connect("changed::view-searchbar", lambda w, k: action_vs.set_state(GLib.Variant.new_boolean(w.get_boolean(k))))
        self.config.connect("changed::view-searchbar", lambda w, k: self.searchbar.set_visible(GLib.Variant.new_boolean(w.get_boolean(k))))
        group.add_action(action_vs)

        action_va = Gio.SimpleAction.new_stateful("view-statusbar", None, GLib.Variant.new_boolean(True))
        action_va.connect("activate", lambda w, k: action_va.set_state(GLib.Variant.new_boolean(not action_va.get_state())))
        action_va.connect("activate", lambda w, k: self.config.set_boolean("view-statusbar", action_va.get_state()))
        action_va.connect("activate", lambda w, k: self.statusbar.set_visible(GLib.Variant.new_boolean(action_va.get_state())))
        self.config.connect("changed::view-statusbar", lambda w, k: action_va.set_state(GLib.Variant.new_boolean(w.get_boolean(k))))
        self.config.connect("changed::view-statusbar", lambda w, k: self.statusbar.set_visible(GLib.Variant.new_boolean(w.get_boolean(k))))
        group.add_action(action_va)

        action_vt = Gio.SimpleAction.new_stateful("view-toolbar", None, GLib.Variant.new_boolean(True))
        action_vt.connect("activate", lambda w, k: action_vt.set_state(GLib.Variant.new_boolean(not action_vt.get_state())))
        action_vt.connect("activate", lambda w, k: self.config.set_boolean("view-toolbar", action_vt.get_state()))
        action_vt.connect("activate", lambda w, k: self.toolbar.set_visible(GLib.Variant.new_boolean(action_vt.get_state())))
        self.config.connect("changed::view-toolbar", lambda w, k: action_vt.set_state(GLib.Variant.new_boolean(w.get_boolean(k))))
        self.config.connect("changed::view-toolbar", lambda w, k: self.toolbar.set_visible(GLib.Variant.new_boolean(w.get_boolean(k))))
        self.config.connect("changed::view-toolbar-style", lambda w, k: self.__cb_config_toolbar_style(w, w.get_string(k)))
        group.add_action(action_vt)


    def __init_facilities(self):
        "Sets up various facilities"

        self.clipboard      = data.Clipboard()
        self.datafile       = io.DataFile(datahandler.Revelation2)
        self.entryclipboard = data.EntryClipboard()
        self.entrystore     = data.EntryStore()
        self.entrysearch    = data.EntrySearch(self.entrystore)
        self.items          = Gtk.IconTheme.get_default()
        self.locktimer      = data.Timer()
        self.undoqueue      = data.UndoQueue()

        self.datafile.connect("changed", lambda w,f: self.__state_file(f))
        self.datafile.connect("content-changed", self.__cb_file_content_changed)
        self.entryclipboard.connect("content-toggled", lambda w,d: self.__state_clipboard(d))
        self.locktimer.connect("ring", self.__cb_file_autolock)
        self.undoqueue.connect("changed", lambda w: self.__state_undo(self.undoqueue.get_undo_action(), self.undoqueue.get_redo_action()))

        self.config.connect("changed::file-autolock-timeout", lambda w, k: self.locktimer.start(60 * w.get_int(k)))
        if self.config.get_boolean("file-autolock"):
            self.locktimer.start(60 * self.config.get_int("file-autolock-timeout"))

        dialog.EVENT_FILTER = self.__cb_event_filter


    def __init_states(self):
        "Sets the initial application state"

        # set window states
        self.window.set_default_size(
            self.config.get_int("view-window-width"),
            self.config.get_int("view-window-height")
        )

        self.window.move(
            self.config.get_int("view-window-position-x"),
            self.config.get_int("view-window-position-y")
        )

        self.hpaned.set_position(
            self.config.get_int("view-pane-position")
        )

        # bind ui widgets to config keys
        bind = {
            "view-passwords"    : "/menubar/menu-view/view-passwords",
            "view-searchbar"    : "/menubar/menu-view/view-searchbar",
            "view-statusbar"    : "/menubar/menu-view/view-statusbar",
            "view-toolbar"      : "/menubar/menu-view/view-toolbar"
        }

        for key in bind.keys():
            self.window.get_action_group("file").lookup_action(key).set_state(self.config.get_value(key))

        self.window.show_all()

        # use some events to restart lock timer
        Gdk.event_handler_set(self.__cb_event_filter)
        self.file_locked = False


        # set some variables
        self.entrysearch.string = ''
        self.entrysearch.type   = None

        # set ui widget states
        self.__state_clipboard(self.entryclipboard.has_contents())
        self.__state_entry([])
        self.__state_file(None)
        self.__state_find(self.searchbar.entry.get_text())
        self.__state_undo(None, None)

        # set states from config
        self.searchbar.set_visible(self.config.get_boolean("view-searchbar"))
        self.statusbar.set_visible(self.config.get_boolean("view-statusbar"))
        self.toolbar.set_visible(self.config.get_boolean("view-toolbar"))
        self.__cb_config_toolbar_style(self.config, self.config.get_string("view-toolbar-style"))

        # give focus to searchbar entry if shown
        if self.searchbar.get_property("visible") == True:
            self.searchbar.entry.grab_focus()


    def __init_ui(self):
        "Sets up the UI"

        # add custom icons path
        _icon_theme = Gtk.IconTheme.get_default()
        if _icon_theme is not None:
            _icon_theme.append_search_path(config.DIR_ICONS)

        # set window icons
        pixbufs = [ self.items.load_icon("info.olasagasti.revelation", size, 0) for size in ( 128, 48, 32, 24, 16) ]
        pixbufs = [ pixbuf for pixbuf in pixbufs if pixbuf != None ]

        if len(pixbufs) > 0:
            Gtk.Window.set_default_icon_list(pixbufs)

        # set up toolbar and menus
        self.set_menubar(self.menubar)

        toolbar = Gtk.Toolbar.new()
        toolbar.set_style(Gtk.ToolbarStyle.BOTH_HORIZ)

        open_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('document-open', Gtk.IconSize.LARGE_TOOLBAR), _('_Open'))
        open_item.connect('clicked', lambda k: self.window.get_action_group("file").lookup_action("file-open").activate())
        open_item.set_tooltip_text(_('Open a file'))
        open_item.set_use_underline(True)
        toolbar.insert(open_item, -1)

        save_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('document-save', Gtk.IconSize.LARGE_TOOLBAR), _('_Save'))
        save_item.connect('clicked', lambda k: self.window.get_action_group("file").lookup_action("file-save").activate())
        save_item.set_tooltip_text(_('Save data to a file'))
        save_item.set_property('is-important', True)
        save_item.set_use_underline(True)
        toolbar.insert(save_item, -1)

        toolbar.insert(Gtk.SeparatorToolItem.new(), -1)

        addentry_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('list-add', Gtk.IconSize.LARGE_TOOLBAR), _('Add Entry'))
        addentry_item.connect('clicked', lambda k: self.window.get_action_group("entry-optional").lookup_action("entry-add").activate())
        addentry_item.set_tooltip_text(_('Create a new entry'))
        addentry_item.set_property('is-important', True)
        addentry_item.set_use_underline(True)
        toolbar.insert(addentry_item, -1)

        addfolder_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('folder-new', Gtk.IconSize.LARGE_TOOLBAR), _('Add folder'))
        addfolder_item.connect('clicked', lambda k: self.window.get_action_group("entry-optional").lookup_action("entry-folder").activate())
        addfolder_item.set_tooltip_text(_('Create a new folder'))
        toolbar.insert(addfolder_item, -1)

        toolbar.insert(Gtk.SeparatorToolItem.new(), -1)

        gotoentry_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('go-jump', Gtk.IconSize.LARGE_TOOLBAR), _('_Go to'))
        gotoentry_item.connect('clicked', lambda k: self.window.get_action_group("dynamic").lookup_action("entry-goto").activate())
        gotoentry_item.set_tooltip_text(_('Go to the selected entries'))
        gotoentry_item.set_property('is-important', True)
        gotoentry_item.set_use_underline(True)
        toolbar.insert(gotoentry_item, -1)

        editentry_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('document-edit', Gtk.IconSize.LARGE_TOOLBAR), _('_Edit'))
        editentry_item.connect('clicked', lambda k: self.window.get_action_group("entry-single").lookup_action("entry-edit").activate())
        editentry_item.set_tooltip_text(_('Edit the selected entry'))
        editentry_item.set_use_underline(True)
        toolbar.insert(editentry_item, -1)

        removeentry_item = Gtk.ToolButton.new(Gtk.Image.new_from_icon_name('edit-delete', Gtk.IconSize.LARGE_TOOLBAR), _('Re_move'))
        removeentry_item.connect('clicked', lambda k: self.window.get_action_group("entry-multiple").lookup_action("entry-remove").activate())
        removeentry_item.set_tooltip_text(_('Remove the selected entries'))
        removeentry_item.set_use_underline(True)
        toolbar.insert(removeentry_item, -1)


        self.toolbar=toolbar
        self.toolbar.connect("popup-context-menu", lambda w,x,y,b: True)
        self.add_toolbar(toolbar, "toolbar", 1)

        self.statusbar = ui.Statusbar()
        self.main_vbox.pack_end(self.statusbar, False, True, 0)

        self.searchbar = ui.Searchbar()
        self.add_toolbar(self.searchbar, "searchbar", 2)

        # set up main application widgets
        self.tree = ui.EntryTree(self.entrystore)
        self.scrolledwindow = ui.ScrolledWindow(self.tree)

        self.entryview = ui.EntryView(self.config, self.clipboard)
        self.entryview.set_halign(Gtk.Align.CENTER)
        self.entryview.set_valign(Gtk.Align.CENTER)
        self.entryview.set_hexpand(True)

        self.hpaned = Gtk.Paned.new(Gtk.Orientation.HORIZONTAL)
        self.hpaned.pack1(self.scrolledwindow, True, True)
        self.hpaned.pack2(self.entryview, True, True)
        self.hpaned.set_border_width(6)
        self.set_contents(self.hpaned)

        # set up drag-and-drop
        uritarget = Gtk.TargetEntry.new("text/uri-list", 0, 0)
        self.window.drag_dest_set(Gtk.DestDefaults.ALL, [uritarget], Gdk.DragAction.COPY | Gdk.DragAction.MOVE | Gdk.DragAction.LINK )
        self.window.connect("drag_data_received", self.__cb_drag_dest)

        self.tree.enable_model_drag_source(Gdk.ModifierType.BUTTON1_MASK, ( ( "revelation/treerow", Gtk.TargetFlags.SAME_APP | Gtk.TargetFlags.SAME_WIDGET, 0), ), Gdk.DragAction.MOVE)
        self.tree.enable_model_drag_dest(( ( "revelation/treerow", Gtk.TargetFlags.SAME_APP | Gtk.TargetFlags.SAME_WIDGET, 0), ), Gdk.DragAction.MOVE)
        self.tree.connect("drag_data_received", self.__cb_tree_drag_received)

        # set up callbacks
        self.searchbar.entry.connect("key-press-event", self.__cb_searchbar_key_press)
        self.searchbar.button_next.connect("clicked", self.__cb_searchbar_button_clicked, data.SEARCH_NEXT)
        self.searchbar.button_prev.connect("clicked", self.__cb_searchbar_button_clicked, data.SEARCH_PREVIOUS)
        self.searchbar.entry.connect("changed", lambda w: self.__state_find(self.searchbar.entry.get_text()))

        self.tree.connect("popup", lambda w,d: self.popup(self.popupmenu, d.button, d.time))
        self.tree.connect("doubleclick", self.__cb_tree_doubleclick)
        self.tree.connect("key-press-event", self.__cb_tree_keypress)
        self.tree.selection.connect("changed", lambda w: self.entryview.display_entry(self.entrystore.get_entry(self.tree.get_active())))
        self.tree.selection.connect("changed", lambda w: self.__state_entry(self.tree.get_selected()))

    def __init_dbus(self):
        app = Gio.Application.get_default
        self.dbus_subscription_id = app().get_dbus_connection().signal_subscribe(None, "org.gnome.ScreenSaver", "ActiveChanged", "/org/gnome/ScreenSaver", None, Gio.DBusSignalFlags.NONE, self.__cb_screensaver_lock)

    ##### STATE HANDLERS #####

    def __save_state(self):
        "Saves the current application state"

        width, height = self.window.get_size()
        self.config.set_int("view-window-width", width)
        self.config.set_int("view-window-height", height)

        x, y = self.window.get_position()
        self.config.set_int("view-window-position-x", x)
        self.config.set_int("view-window-position-y", y)

        self.config.set_int("view-pane-position", self.hpaned.get_position())
        self.config.sync()


    def __state_clipboard(self, has_contents):
        "Sets states based on the clipboard contents"

        self.window.get_action_group("dynamic").lookup_action("clip-paste").set_enabled(has_contents)

    def __state_entry(self, iters):
        "Sets states for entry-dependant ui items"

        # widget sensitivity based on number of entries
        for action in self.window.get_action_group("entry-multiple").list_actions():
            self.window.get_action_group("entry-multiple").lookup_action(action).set_enabled(len(iters) > 0)
        for action in self.window.get_action_group("entry-single").list_actions():
            self.window.get_action_group("entry-single").lookup_action(action).set_enabled(len(iters) == 1)
        for action in self.window.get_action_group("entry-optional").list_actions():
            self.window.get_action_group("entry-optional").lookup_action(action).set_enabled(len(iters) < 2)


        # copy password sensitivity
        s = False

        for iter in iters:
            e = self.entrystore.get_entry(iter)

            for f in e.fields:
                if f.datatype == entry.DATATYPE_PASSWORD and f.value != "":
                    s = True

        self.window.get_action_group("entry-multiple").lookup_action("clip-chain").set_enabled(s)


        # goto sensitivity
        try:
            for iter in iters:
                e = self.entrystore.get_entry(iter)

                if (e.id == "folder"):
                    continue

                if self.config.get_string("launcher-%s" % e.id) not in ( "", None ):
                    s = True
                    break

            else:
                s = False

        except config.ConfigError:
            s = False

        self.window.get_action_group("dynamic").lookup_action("entry-goto").set_enabled(s)


    def __state_file(self, file):
        "Sets states based on file"

        for action in self.window.get_action_group("file-exists").list_actions():
            self.window.get_action_group("file-exists").lookup_action(action).set_enabled(file is not None)

        if file is not None:
            self.window.set_title(os.path.basename(file))

            if io.file_is_local(file):
                os.chdir(os.path.dirname(file))

        else:
            self.window.set_title('[' + _('New file') + ']')


    def __state_find(self, string):
        "Sets states based on the current search string"

        for action in self.window.get_action_group("find").list_actions():
            self.window.get_action_group("find").lookup_action(action).set_enabled(string != "")


    def __state_undo(self, undoaction, redoaction):
        "Sets states based on undoqueue actions"

        if undoaction is None:
            s, l = False, _('_Undo')

        else:
            s, l = True, _('_Undo %s') % undoaction[1].lower()

        action = self.window.get_action_group("dynamic").lookup_action("undo")
        action.set_enabled(s)
        # TODO action.set_property("label", l)


        if redoaction is None:
            s, l = False, _('_Redo')

        else:
            s, l = True, _('_Redo %s') % redoaction[1].lower()

        action = self.window.get_action_group("dynamic").lookup_action("redo")
        action.set_enabled(s)
        # TODO action.set_property("label", l)




    ##### MISC CALLBACKS #####

    def __cb_clip_copy(self, widget, data = None):
        "Handles copying to the clipboard"

        focuswidget = self.window.get_focus()

        if focuswidget is self.tree:
            self.clip_copy(self.tree.get_selected())

        elif isinstance(focuswidget, Gtk.Label) or isinstance(focuswidget, Gtk.Entry):
            focuswidget.emit("copy-clipboard")


    def __cb_clip_cut(self, widget, data = None):
        "Handles cutting to clipboard"

        focuswidget = self.window.get_focus()

        if focuswidget is self.tree:
            self.clip_cut(self.tree.get_selected())

        elif isinstance(focuswidget, Gtk.Entry):
            focuswidget.emit("cut-clipboard")


    def __cb_clip_paste(self, widget, data = None):
        "Handles pasting from clipboard"

        focuswidget = self.window.get_focus()

        if focuswidget is self.tree:
            self.clip_paste(self.entryclipboard.get(), self.tree.get_active())

        elif isinstance(focuswidget, Gtk.Entry):
            focuswidget.emit("paste-clipboard")


    def __cb_drag_dest(self, widget, context, x, y, seldata, info, time, userdata = None):
        "Handles file drops"

        if seldata.data is None:
            return

        files = [ file.strip() for file in seldata.data.split("\n") if file.strip() != "" ]

        if len(files) > 0:
            self.file_open(files[0])


    def __cb_event_filter(self, event):
        "Event filter for gdk window"

        if event.type in (Gdk.EventType.KEY_PRESS, Gdk.EventType.BUTTON_PRESS, Gdk.EventType.MOTION_NOTIFY):
            self.locktimer.reset()

        Gtk.main_do_event(event)
        return Gdk.FilterReturn.CONTINUE


    def __cb_exception(self, type, value, trace):
        "Callback for unhandled exceptions"

        if type == KeyboardInterrupt:
            sys.exit(1)

        traceback = util.trace_exception(type, value, trace)
        sys.stderr.write(traceback)

        if dialog.Exception(self.window, traceback).run() == True:
            Gtk.main()

        else:
            sys.exit(1)


    def __cb_file_content_changed(self, widget, file):
        "Callback for changed file"

        try:
            if dialog.FileChanged(self.window, file).run() == True:
                self.file_open(self.datafile.get_file(), self.datafile.get_password())

        except dialog.CancelError:
            self.statusbar.set_status(_('Open cancelled'))


    def __cb_file_autolock(self, widget, data = None):
        "Callback for locking the file"

        if self.config.get_boolean("file-autolock") == True:
            self.file_lock()


    def __cb_screensaver_lock(self, connection, unique_name, object_path, interface, signal, state):
        if state[0] is True and self.config.get_boolean("file-autolock") == True:
            self.file_lock()

    def __cb_quit(self, widget, data = None):
        "Callback for quit"

        if self.quit() == False:
            return True

        else:
            return False

    def __cb_close(self, widget, data = None):
        "Callback for Close"

        if self.file_close() == False:
            return True

        else:
            return False

    def __cb_searchbar_button_clicked(self, widget, direction = data.SEARCH_NEXT):
        "Callback for searchbar button clicks"

        self.__entry_find(self, self.searchbar.entry.get_text(), self.searchbar.dropdown.get_active_type(), direction)
        self.searchbar.entry.select_region(0, -1)


    def __cb_searchbar_key_press(self, widget, data):
        "Callback for searchbar key presses"

        # escape
        if data.keyval == Gdk.KEY_Escape:
            context = widget.get_style_context()
            context.remove_class(Gtk.STYLE_CLASS_ERROR)
            self.config.set_boolean("view-searchbar", False)


    def __cb_tree_doubleclick(self, widget, iter):
        "Handles doubleclicks on the tree"

        if self.config.get_string("behavior-doubleclick") == "edit":
            self.entry_edit(iter)

        elif self.config.get_string("behavior-doubleclick") == "copy":
            self.clip_chain(self.entrystore.get_entry(iter))

        else:
            self.entry_goto((iter,))


    def __cb_tree_drag_received(self, tree, context, x, y, seldata, info, time):
        "Callback for drag drops on the treeview"

        # get source and destination data
        sourceiters = self.entrystore.filter_parents(self.tree.get_selected())
        destrow = self.tree.get_dest_row_at_pos(x, y)

        if destrow is None:
            destpath = ( self.entrystore.iter_n_children(None) - 1, )
            pos = Gtk.TreeViewDropPosition.AFTER

        else:
            destpath, pos = destrow

        destiter = self.entrystore.get_iter(destpath)
        destpath = self.entrystore.get_path(destiter)

        # avoid drops to current iter or descentants
        for sourceiter in sourceiters:
            sourcepath = self.entrystore.get_path(sourceiter)

            if self.entrystore.is_ancestor(sourceiter, destiter) == True or sourcepath == destpath:
                context.finish(False, False, time)
                return

            elif pos == Gtk.TreeViewDropPosition.BEFORE and sourcepath[:-1] == destpath[:-1] and sourcepath[-1] == destpath[-1] - 1:
                context.finish(False, False, time)
                return

            elif pos == Gtk.TreeViewDropPosition.AFTER and sourcepath[:-1] == destpath[:-1] and sourcepath[-1] == destpath[-1] + 1:
                context.finish(False, False, time)
                return


        # move the entries
        if pos in ( Gtk.TreeViewDropPosition.INTO_OR_BEFORE, Gtk.TreeViewDropPosition.INTO_OR_AFTER):
            parent = destiter
            sibling = None

        elif pos == Gtk.TreeViewDropPosition.BEFORE:
            parent = self.entrystore.iter_parent(destiter)
            sibling = destiter

        elif pos == Gtk.TreeViewDropPosition.AFTER:
            parent = self.entrystore.iter_parent(destiter)

            sibpath = list(destpath)
            sibpath[-1] += 1
            sibling = self.entrystore.get_iter(sibpath)

        self.entry_move(sourceiters, parent, sibling)

        context.finish(False, False, time)


    def __cb_tree_keypress(self, widget, data = None):
        "Handles key presses for the tree"

        # return
        if data.keyval == Gdk.KEY_Return:
            self.entry_edit(self.tree.get_active())

        # insert
        elif data.keyval == Gdk.KEY_Insert:
            self.entry_add(None, self.tree.get_active())

        # delete
        elif data.keyval == Gdk.KEY_Delete:
            self.entry_remove(self.tree.get_selected())



    ##### CONFIG CALLBACKS #####

    def __cb_config_toolbar_style(self, config, value, data = None):
        "Config callback for setting toolbar style"

        if value == "both":
            self.toolbar.set_style(Gtk.ToolbarStyle.BOTH)

        elif value == "both-horiz":
            self.toolbar.set_style(Gtk.ToolbarStyle.BOTH_HORIZ)

        elif value == "icons":
            self.toolbar.set_style(Gtk.ToolbarStyle.ICONS)

        elif value == "text":
            self.toolbar.set_style(Gtk.ToolbarStyle.TEXT)

        else:
            self.toolbar.unset_style()


    #### UNDO / REDO CALLBACKS #####

    def __cb_redo_add(self, name, actiondata):
        "Redoes an add action"

        path, e = actiondata
        parent = self.entrystore.get_iter(path[:-1])
        sibling = self.entrystore.get_iter(path)

        iter = self.entrystore.add_entry(e, parent, sibling)
        self.tree.select(iter)


    def __cb_redo_edit(self, name, actiondata):
        "Redoes an edit action"

        path, preentry, postentry = actiondata
        iter = self.entrystore.get_iter(path)

        self.entrystore.update_entry(iter, postentry)
        self.tree.select(iter)


    def __cb_redo_import(self, name, actiondata):
        "Redoes an import action"

        paths, entrystore = actiondata
        self.entrystore.import_entry(entrystore, None)


    def __cb_redo_move(self, name, actiondata):
        "Redoes a move action"

        newiters = []

        for prepath, postpath in actiondata:
            prepath, postpath = list(prepath), list(postpath)

            # adjust path if necessary
            if len(prepath) <= len(postpath):
                if prepath[:-1] == postpath[:len(prepath) - 1]:
                    if prepath[-1] <= postpath[len(prepath) - 1]:
                        postpath[len(prepath) - 1] += 1

            newiter = self.entrystore.move_entry(
                self.entrystore.get_iter(prepath),
                self.entrystore.get_iter(postpath[:-1]),
                self.entrystore.get_iter(postpath)
            )

            newiters.append(newiter)

        if len(newiters) > 0:
            self.tree.select(newiters[0])


    def __cb_redo_paste(self, name, actiondata):
        "Redoes a paste action"

        entrystore, parentpath, paths = actiondata
        iters = self.entrystore.import_entry(entrystore, None, self.entrystore.get_iter(parentpath))

        if len(iters) > 0:
            self.tree.select(iters[0])


    def __cb_redo_remove(self, name, actiondata):
        "Redoes a remove action"

        iters = []
        for path, entrystore in actiondata:
            iters.append(self.entrystore.get_iter(path))

        for iter in iters:
            self.entrystore.remove_entry(iter)

        self.tree.unselect_all()


    def __cb_undo_add(self, name, actiondata):
        "Undoes an add action"

        path, e = actiondata

        self.entrystore.remove_entry(self.entrystore.get_iter(path))
        self.tree.unselect_all()


    def __cb_undo_edit(self, name, actiondata):
        "Undoes an edit action"

        path, preentry, postentry = actiondata
        iter = self.entrystore.get_iter(path)

        self.entrystore.update_entry(iter, preentry)
        self.tree.select(iter)


    def __cb_undo_import(self, name, actiondata):
        "Undoes an import action"

        paths, entrystore = actiondata
        iters = [ self.entrystore.get_iter(path) for path in paths ]

        for iter in iters:
            self.entrystore.remove_entry(iter)

        self.tree.unselect_all()


    def __cb_undo_move(self, name, actiondata):
        "Undoes a move action"

        actiondata = actiondata[:]
        actiondata.reverse()

        newiters = []

        for prepath, postpath in actiondata:
            prepath, postpath = list(prepath), list(postpath)

            # adjust path if necessary
            if len(postpath) <= len(prepath):
                if postpath[:-1] == prepath[:len(postpath) - 1]:
                    if postpath[-1] <= prepath[len(postpath) - 1]:
                        prepath[len(postpath) - 1] += 1

            newiter = self.entrystore.move_entry(
                self.entrystore.get_iter(postpath),
                self.entrystore.get_iter(prepath[:-1]),
                self.entrystore.get_iter(prepath)
            )

            newiters.append(newiter)

        if len(newiters) > 0:
            self.tree.select(newiters[-1])


    def __cb_undo_paste(self, name, actiondata):
        "Undoes a paste action"

        entrystore, parentpath, paths = actiondata
        iters = [ self.entrystore.get_iter(path) for path in paths ]

        for iter in iters:
            self.entrystore.remove_entry(iter)

        self.tree.unselect_all()


    def __cb_undo_remove(self, name, actiondata):
        "Undoes a remove action"

        iters = []
        for path, entrystore in actiondata:
            parent = self.entrystore.get_iter(path[:-1])
            sibling = self.entrystore.get_iter(path)

            iter = self.entrystore.import_entry(entrystore, entrystore.iter_nth_child(None, 0), parent, sibling)
            iters.append(iter)

        self.tree.select(iters[0])



    ##### PRIVATE METHODS #####

    def __entry_find(self, parent, string, entrytype, direction = data.SEARCH_NEXT):
        "Searches for an entry"

        match = self.entrysearch.find(string, entrytype, self.tree.get_active(), direction)
        context = self.searchbar.entry.get_style_context()

        if match != None:
            self.tree.select(match)
            self.statusbar.set_status(_('Match found for “%s”') % string)
            context.remove_class(Gtk.STYLE_CLASS_ERROR)

        else:
            self.statusbar.set_status(_('No match found for “%s”') % string)
            context.add_class(Gtk.STYLE_CLASS_ERROR)


    def __file_autosave(self):
        "Autosaves the current file if needed"

        try:
            if self.datafile.get_file() is None or self.datafile.get_password() is None:
                return

            if self.config.get_boolean("file-autosave") == False:
                return

            self.datafile.save(self.entrystore, self.datafile.get_file(), self.datafile.get_password())
            self.entrystore.changed = False

        except IOError:
            pass


    def __file_load(self, file, password, datafile = None):
        "Loads data from a data file into an entrystore"

        # We may need to change the datahandler
        old_handler = None
        result = None

        try:
            if datafile is None:
                datafile = self.datafile

                # Because there are two fileversion we need to check if we are really dealing
                # with version two. If we aren't chances are high, that we are
                # dealing with version one. In this case we use the version one
                # handler and save the file as version two if it is changed, to
                # allow seemless upgrades.
                if datafile.get_handler().detect(io.file_read(file)) != True:
                    # Store the datahandler to be reset later on
                    old_handler = datafile.get_handler()
                    # Load the revelation fileversion one handler
                    datafile.set_handler(datahandler.Revelation)
                    dialog.Info(self.window,_('Old file format'), _('Revelation detected that \'%s\' file has the old and actually non-secure file format. It is strongly recommended to save this file with the new format. Revelation will do it automatically if you press save after opening the file.') % file).run()

            while True:
                try:
                    result = datafile.load(file, password, lambda: dialog.PasswordOpen(self.window, os.path.basename(file)).run())
                    break

                except datahandler.PasswordError:
                    dialog.Error(self.window, _('Incorrect password'), _('The password you entered for the file \'%s\' was not correct.') % file).run()

        except datahandler.FormatError:
            self.statusbar.set_status(_('Open failed'))
            dialog.Error(self.window, _('Invalid file format'), _('The file \'%s\' contains invalid data.') % file).run()

        except ( datahandler.DataError, entry.EntryTypeError, entry.EntryFieldError ):
            self.statusbar.set_status(_('Open failed'))
            dialog.Error(self.window, _('Unknown data'), _('The file \'%s\' contains unknown data. It may have been created by a newer version of Revelation.') % file).run()

        except datahandler.VersionError:
            self.statusbar.set_status(_('Open failed'))
            dialog.Error(self.window, _('Unknown data version'), _('The file \'%s\' has a future version number, please upgrade Revelation to open it.') % file).run()

        except datahandler.DetectError:
            self.statusbar.set_status(_('Open failed'))
            dialog.Error(self.window, _('Unable to detect filetype'), _('The file type of the file \'%s\' could not be automatically detected. Try specifying the file type manually.')% file).run()

        except IOError:
            self.statusbar.set_status(_('Open failed'))
            dialog.Error(self.window, _('Unable to open file'), _('The file \'%s\' could not be opened. Make sure that the file exists, and that you have permissions to open it.') % file).run()

        # If we switched the datahandlers before we need to switch back to the
        # version2 handler here, to ensure a seemless version upgrade on save
        if old_handler != None:
            datafile.set_handler(old_handler.__class__)

        return result


    def __get_common_usernames(self, e = None):
        "Returns a list of possibly relevant usernames"

        ulist = []

        if e is not None and e.has_field(entry.UsernameField):
            ulist.append(e[entry.UsernameField])

        ulist.append(pwd.getpwuid(os.getuid())[0])
        ulist.extend(self.entrystore.get_popular_values(entry.UsernameField, 3))

        ulist = list({}.fromkeys(ulist).keys())
        ulist.sort()

        return ulist



    ##### PUBLIC METHODS #####

    def about(self):
        "Displays the about dialog"

        dialog.run_unique(dialog.About, self)


    def clip_chain(self, e):
        "Copies all passwords from an entry as a chain"

        if e is None:
            return

        secrets = [ field.value for field in e.fields if field.datatype == entry.DATATYPE_PASSWORD and field.value != "" ]

        if self.config.get_boolean("clipboard-chain-username") == True and len(secrets) > 0 and e.has_field(entry.UsernameField) and e[entry.UsernameField] != "":
            secrets.insert(0, e[entry.UsernameField])

        if len(secrets) == 0:
            self.statusbar.set_status(_('Entry has no password to copy'))

        else:
            self.clipboard.set(secrets, True)
            self.statusbar.set_status(_('Password copied to clipboard'))


    def clip_copy(self, iters):
        "Copies entries to the clipboard"

        self.entryclipboard.set(self.entrystore, iters)
        self.statusbar.set_status(_('Entries copied'))


    def clip_cut(self, iters):
        "Cuts entries to the clipboard"

        iters = self.entrystore.filter_parents(iters)
        self.entryclipboard.set(self.entrystore, iters)

        # store undo data (need paths)
        undoactions = []
        for iter in iters:
            undostore = data.EntryStore()
            undostore.import_entry(self.entrystore, iter)
            path = self.entrystore.get_path(iter)
            undoactions.append( ( path, undostore ) )

        # remove data
        for iter in iters:
            self.entrystore.remove_entry(iter)

        self.undoqueue.add_action(
            _('Cut entries'), self.__cb_undo_remove, self.__cb_redo_remove,
            undoactions
        )

        self.__file_autosave()

        self.tree.unselect_all()
        self.statusbar.set_status(_('Entries cut'))


    def clip_paste(self, entrystore, parent):
        "Pastes entries from the clipboard"

        if entrystore is None:
            return

        parent = self.tree.get_active()
        iters = self.entrystore.import_entry(entrystore, None, parent)

        paths = [ self.entrystore.get_path(iter) for iter in iters ]

        self.undoqueue.add_action(
            _('Paste entries'), self.__cb_undo_paste, self.__cb_redo_paste,
            ( entrystore, self.entrystore.get_path(parent), paths )
        )

        if len(iters) > 0:
            self.tree.select(iters[0])

        self.statusbar.set_status(_('Entries pasted'))


    def entry_add(self, e = None, parent = None, sibling = None):
        "Adds an entry"

        try:
            if e is None:
                d = dialog.EntryEdit(self.window, _('Add Entry'), None, self.config, self.clipboard)
                d.set_fieldwidget_data(entry.UsernameField, self.__get_common_usernames())
                e = d.run()

            iter = self.entrystore.add_entry(e, parent, sibling)

            self.undoqueue.add_action(
                _('Add entry'), self.__cb_undo_add, self.__cb_redo_add,
                ( self.entrystore.get_path(iter), e.copy() )
            )

            self.__file_autosave()
            self.tree.select(iter)
            self.statusbar.set_status(_('Entry added'))

        except dialog.CancelError:
            self.statusbar.set_status(_('Add entry cancelled'))


    def entry_edit(self, iter):
        "Edits an entry"

        try:
            if iter is None:
                return

            e = self.entrystore.get_entry(iter)

            if type(e) == entry.FolderEntry:
                d = dialog.FolderEdit(self.window, _('Edit Folder'), e)

            else:
                d = dialog.EntryEdit(self.window, _('Edit Entry'), e, self.config, self.clipboard)
                d.set_fieldwidget_data(entry.UsernameField, self.__get_common_usernames(e))


            n = d.run()
            self.entrystore.update_entry(iter, n)
            self.tree.select(iter)

            self.undoqueue.add_action(
                _('Update entry'), self.__cb_undo_edit, self.__cb_redo_edit,
                ( self.entrystore.get_path(iter), e.copy(), n.copy() )
            )

            self.__file_autosave()
            self.statusbar.set_status(_('Entry updated'))

        except dialog.CancelError:
            self.statusbar.set_status(_('Edit entry cancelled'))


    def entry_find(self):
        "Searches for an entry"

        self.config.set_boolean("view-searchbar", True)
        self.searchbar.entry.select_region(0, -1)
        self.searchbar.entry.grab_focus()


    def entry_folder(self, e = None, parent = None, sibling = None):
        "Adds a folder"

        try:
            if e is None:
                e = dialog.FolderEdit(self.window, _('Add folder')).run()

            iter = self.entrystore.add_entry(e, parent, sibling)

            self.undoqueue.add_action(
                _('Add folder'), self.__cb_undo_add, self.__cb_redo_add,
                ( self.entrystore.get_path(iter), e.copy() )
            )

            self.__file_autosave()
            self.tree.select(iter)
            self.statusbar.set_status(_('Folder added'))

        except dialog.CancelError:
            self.statusbar.set_status(_('Add folder cancelled'))


    def entry_goto(self, iters):
        "Goes to an entry"

        for iter in iters:
            try:

                # get goto data for entry
                e = self.entrystore.get_entry(iter)
                command = self.config.get_string("launcher-%s" % e.id)

                if command in ( "", None ):
                    self.statusbar.set_status(_('No goto command found for %s entries') % e.typename)
                    return

                subst = {}
                for field in e.fields:
                    subst[field.symbol] = field.value

                # copy passwords to clipboard
                chain = []

                for field in e.fields:
                    if field.datatype == entry.DATATYPE_PASSWORD and field.value != "":
                        chain.append(field.value)

                if self.config.get_boolean("clipboard-chain-username") == True and len(chain) > 0 and e.has_field(entry.UsernameField) == True and e[entry.UsernameField] != "" and "%" + entry.UsernameField.symbol not in command:
                    chain.insert(0, e[entry.UsernameField])

                self.clipboard.set(chain, True)

                # generate and run goto command
                command = util.parse_subst(command, subst)
                util.execute_child(command)

                self.statusbar.set_status(_('Entry opened'))

            except ( util.SubstFormatError, config.ConfigError ):
                dialog.Error(self.window, _('Invalid goto command format'), _('The goto command for \'%s\' entries is invalid, please correct it in the preferences.') % e.typename).run()

            except util.SubstValueError:
                dialog.Error(self.window, _('Missing entry data'), _('The entry \'%s\' does not have all the data required to open it.') % e.name).run()


    def entry_move(self, sourceiters, parent = None, sibling = None):
        "Moves a set of entries"

        if type(sourceiters) != list:
            sourceiters = [ sourceiters ]

        newiters = []
        undoactions = []

        for sourceiter in sourceiters:
            sourcepath = self.entrystore.get_path(sourceiter)
            newiter = self.entrystore.move_entry(sourceiter, parent, sibling)
            newpath = self.entrystore.get_path(newiter)

            undoactions.append( ( sourcepath, newpath ) )
            newiters.append(newiter)

        self.undoqueue.add_action(
            _('Move entry'), self.__cb_undo_move, self.__cb_redo_move,
            undoactions
        )

        if len(newiters) > 0:
            self.tree.select(newiters[0])

        self.__file_autosave()
        self.statusbar.set_status(_('Entries moved'))


    def entry_remove(self, iters):
        "Removes the selected entries"

        try:
            if len(iters) == 0:
                return

            entries = [ self.entrystore.get_entry(iter) for iter in iters ]
            dialog.EntryRemove(self.window, entries).run()
            iters = self.entrystore.filter_parents(iters)

            # store undo data (need paths)
            undoactions = []
            for iter in iters:
                undostore = data.EntryStore()
                undostore.import_entry(self.entrystore, iter)
                path = self.entrystore.get_path(iter)
                undoactions.append( ( path, undostore ) )

            # remove data
            for iter in iters:
                self.entrystore.remove_entry(iter)

            self.undoqueue.add_action(
                _('Remove entry'), self.__cb_undo_remove, self.__cb_redo_remove,
                undoactions
            )

            self.tree.unselect_all()
            self.__file_autosave()
            self.statusbar.set_status(_('Entries removed'))

        except dialog.CancelError:
            self.statusbar.set_status(_('Entry removal cancelled'))


    def file_change_password(self, password = None):
        "Changes the password of the current data file"

        try:
            if password is None:
                password = dialog.PasswordChange(self.window, self.datafile.get_password()).run()

            self.datafile.set_password(password)
            self.entrystore.changed = True

            self.__file_autosave()
            self.statusbar.set_status(_('Password changed'))

        except dialog.CancelError:
            self.statusbar.set_status(_('Password change cancelled'))

    def file_close(self):
        "Closes the current file"

        try:
            if self.entrystore.changed == True and dialog.FileChangesClose(self.window).run() == True:
                if self.file_save(self.datafile.get_file(), self.datafile.get_password()) == False:
                    raise dialog.CancelError

            self.clipboard.clear()
            self.entryclipboard.clear()
            self.entrystore.clear()
            self.undoqueue.clear()
            self.statusbar.set_status(_('Closed file %s') % self.datafile.get_file())
            self.datafile.close()

            return True

        except dialog.CancelError:
            self.statusbar.set_status(_('Close file cancelled'))
            return False

    def file_export(self):
        "Exports data to a foreign file format"

        try:
            file, handler = dialog.ExportFileSelector(self.window).run()
            datafile = io.DataFile(handler)

            if datafile.get_handler().encryption == True:
                password = dialog.PasswordSave(self.window, file).run()

            else:
                dialog.FileSaveInsecure(self.window).run()
                password = None

            datafile.save(self.entrystore, file, password)
            self.statusbar.set_status(_('Data exported to %s') % datafile.get_file())

        except dialog.CancelError:
            self.statusbar.set_status(_('Export cancelled'))

        except IOError:
            dialog.Error(self.window, _('Unable to write to file'), _('The file \'%s\' could not be opened for writing. Make sure that you have the proper permissions to write to it.') % file).run()
            self.statusbar.set_status(_('Export failed'))


    def file_import(self):
        "Imports data from a foreign file"

        try:
            file, handler = dialog.ImportFileSelector(self.window).run()
            datafile = io.DataFile(handler)
            entrystore = self.__file_load(file, None, datafile)

            if entrystore is not None:
                newiters = self.entrystore.import_entry(entrystore, None)
                paths = [ self.entrystore.get_path(iter) for iter in newiters ]

                self.undoqueue.add_action(
                    _('Import data'), self.__cb_undo_import, self.__cb_redo_import,
                    ( paths, entrystore )
                )

                self.statusbar.set_status(_('Data imported from %s') % datafile.get_file())

            self.__file_autosave()

        except dialog.CancelError:
            self.statusbar.set_status(_('Import cancelled'))


    def file_lock(self):
        "Locks the current file"

        if self.datafile.get_file() is None:
            return

        if self.file_locked is True:
            return

        password = self.datafile.get_password()

        if password is None:
            return

        self.locktimer.stop()
        app = Gio.Application.get_default
        app().get_dbus_connection().signal_unsubscribe(self.dbus_subscription_id)


        # TODO can this be done more elegantly?
        transients = [ window for window in Gtk.Window.list_toplevels() if window.get_transient_for() == self ]

        # store current state
        activeiter = self.tree.get_active()
        oldtitle = self.get_title()

        # clear application contents
        self.tree.set_model(None)
        self.entryview.clear()
        self.window.set_title('[' + _('Locked') + ']')
        self.statusbar.set_status(_('File locked'))
        self.file_locked = True;

        # hide any dialogs
        for window in transients:
            window.hide()

        # lock file
        try:
            d = dialog.PasswordLock(self.window, password)

            if self.entrystore.changed == True:
                l = ui.ImageLabel(_('Quit disabled due to unsaved changes'), "dialog-warning")
                d.contents.pack_start(l, True, True, 0)
                d.get_widget_for_response(Gtk.ResponseType.CANCEL).set_sensitive(False)

            d.run()

        except dialog.CancelError:
            self.quit()

        # unlock the file and restore state
        self.tree.set_model(self.entrystore)
        self.tree.select(activeiter)
        self.window.set_title(oldtitle)
        self.statusbar.set_status(_('File unlocked'))
        self.file_locked = False;

        for window in transients:
            window.show()

        self.locktimer.start(self.config.get_int("file-autolock-timeout") * 60)
        app = Gio.Application.get_default
        self.dbus_subscription_id = app().get_dbus_connection().signal_subscribe(None, "org.gnome.ScreenSaver", "ActiveChanged", "/org/gnome/ScreenSaver", None, Gio.DBusSignalFlags.NONE, self.__cb_screensaver_lock)



    def file_new(self):
        "Opens a new file"

        try:
            if self.entrystore.changed == True and dialog.FileChangesNew(self.window).run() == True:
                if self.file_save(self.datafile.get_file(), self.datafile.get_password()) == False:
                    raise dialog.CancelError

            self.entrystore.clear()
            self.datafile.close()
            self.undoqueue.clear()
            self.statusbar.set_status(_('New file created'))

        except dialog.CancelError:
            self.statusbar.set_status(_('New file cancelled'))


    def file_open(self, file = None, password = None):
        "Opens a data file"

        try:
            if self.entrystore.changed == True and dialog.FileChangesOpen(self.window).run() == True:
                if self.file_save(self.datafile.get_file(), self.datafile.get_password()) == False:
                    raise dialog.CancelError

            if file is None:
                file = dialog.OpenFileSelector(self.window).run()

            entrystore = self.__file_load(file, password)

            if entrystore is None:
                return

            self.entrystore.clear()
            self.entrystore.import_entry(entrystore, None)
            self.entrystore.changed = False
            self.undoqueue.clear()

            self.file_locked = False;
            self.locktimer.start(60 * self.config.get_int("file-autolock-timeout"))
            self.statusbar.set_status(_('Opened file %s') % self.datafile.get_file())

        except dialog.CancelError:
            self.statusbar.set_status(_('Open cancelled'))


    def file_save(self, file = None, password = None):
        "Saves data to a file"

        try:
            if file is None:
                file = dialog.SaveFileSelector(self.window).run()

            if password is None:
                password = dialog.PasswordSave(self.window, file).run()

            self.datafile.save(self.entrystore, file, password)
            self.entrystore.changed = False
            self.statusbar.set_status(_('Data saved to file %s') % file)

            return True

        except dialog.CancelError:
            self.statusbar.set_status(_('Save cancelled'))
            return False

        except IOError:
            dialog.Error(self.window, _('Unable to save file'), _('The file \'%s\' could not be opened for writing. Make sure that you have the proper permissions to write to it.') % file).run()
            self.statusbar.set_status(_('Save failed'))
            return False


    def prefs(self):
        "Displays the application preferences"

        dialog.run_unique(Preferences, self.window, self.config)


    def pwcheck(self):
        "Displays the password checking dialog"

        dialog.run_unique(dialog.PasswordChecker, self.window, self.config, self.clipboard)


    def pwgen(self):
        "Displays the password generator dialog"

        dialog.run_unique(dialog.PasswordGenerator, self.window, self.config, self.clipboard)


    def quit(self):
        "Quits the application"

        try:
            if self.entrystore.changed == True and dialog.FileChangesQuit(self.window).run() == True:
                if self.file_save(self.datafile.get_file(), self.datafile.get_password()) == False:
                    raise dialog.CancelError

            self.clipboard.clear()
            self.entryclipboard.clear()

            self.__save_state()

            Gtk.Application.quit(self)
            if sys.exc_info()[1]:
                # avoid raising an additional exception
                os._exit(0)
            else:
                sys.exit(0)
            return True

        except dialog.CancelError:
            self.statusbar.set_status(_('Quit cancelled'))
            return False


    def redo(self):
        "Redoes the previous action"

        action = self.undoqueue.get_redo_action()

        if action is None:
            return

        self.undoqueue.redo()
        self.statusbar.set_status(_('%s redone') % action[1])
        self.__file_autosave()


    def undo(self):
        "Undoes the previous action"

        action = self.undoqueue.get_undo_action()

        if action is None:
            return

        self.undoqueue.undo()
        self.statusbar.set_status(_('%s undone') % action[1])
        self.__file_autosave()



class Preferences(dialog.Utility):
    "A preference dialog"

    def __init__(self, parent, cfg=None):
        dialog.Utility.__init__(self, parent, _('Preferences'))
        self.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        self.set_default_response(Gtk.ResponseType.CLOSE)
        self.config = cfg
        self.set_modal(False)

        self.notebook = ui.Notebook()
        self.vbox.pack_start(self.notebook, True, True, 0)

        self.page_general = self.notebook.create_page(_('General'))
        self.__init_section_files(self.page_general)
        self.__init_section_password(self.page_general)

        self.page_interface = self.notebook.create_page(_('Interface'))
        self.__init_section_doubleclick(self.page_interface)
        self.__init_section_toolbar(self.page_interface)

        self.page_gotocmd = self.notebook.create_page(_('Goto Commands'))
        self.notebook.get_tab_label(self.page_gotocmd).set_hexpand(True)
        self.__init_section_gotocmd(self.page_gotocmd)

        self.connect("response", lambda w,d: self.destroy())


    def __init_section_doubleclick(self, page):
        "Sets up the doubleclick section"

        self.section_doubleclick = page.add_section(_('Doubleclick Action'))

        # radio-button for go to
        self.radio_doubleclick_goto = Gtk.RadioButton.new_with_label_from_widget(None, _('Go to account, if possible'))
        self.radio_doubleclick_goto.connect("toggled", lambda w: w.get_active() and self.config.set_string("behavior-doubleclick", "goto"))

        self.radio_doubleclick_goto.set_tooltip_text(_('Go to the account (open in external application) on doubleclick, if required data is filled in'))
        self.section_doubleclick.append_widget(None, self.radio_doubleclick_goto)

        # radio-button for edit
        self.radio_doubleclick_edit = Gtk.RadioButton.new_with_label_from_widget(self.radio_doubleclick_goto, label=_('Edit account'))
        self.radio_doubleclick_edit.connect("toggled", lambda w: w.get_active() and self.config.set_string("behavior-doubleclick", "edit"))

        self.radio_doubleclick_edit.set_tooltip_text(_('Edit the account on doubleclick'))
        self.section_doubleclick.append_widget(None, self.radio_doubleclick_edit)

        # radio-button for copy
        self.radio_doubleclick_copy = Gtk.RadioButton.new_with_label_from_widget(self.radio_doubleclick_goto, label=_('Copy password to clipboard'))
        self.radio_doubleclick_copy.connect("toggled", lambda w: w.get_active() and self.config.set_string("behavior-doubleclick", "copy"))

        self.radio_doubleclick_copy.set_tooltip_text(_('Copy the account password to clipboard on doubleclick'))
        self.section_doubleclick.append_widget(None, self.radio_doubleclick_copy)

        {"goto":self.radio_doubleclick_goto,
         "edit":self.radio_doubleclick_edit,
         "copy":self.radio_doubleclick_copy}[self.config.get_string("behavior-doubleclick")].set_active(True)


    def __init_section_files(self, page):
        "Sets up the files section"

        self.section_files = page.add_section(_('Files'))

        # checkbutton and file button for autoloading a file
        self.check_autoload = Gtk.CheckButton(label=_('Open file on startup:'))
        self.config.bind("file-autoload", self.check_autoload, "active", Gio.SettingsBindFlags.DEFAULT)

        self.check_autoload.connect("toggled", lambda w: self.button_autoload_file.set_sensitive(w.get_active()))
        self.check_autoload.set_tooltip_text(_('When enabled, this file will be opened when the program is started'))

        self.button_autoload_file = Gtk.FileChooserButton(title=_('Select File to Automatically Open'))
        if self.config.get_boolean("file-autoload"):
            self.button_autoload_file.set_filename(self.config.get_string("file-autoload-file"))
        self.button_autoload_file.connect('file-set', lambda w: self.config.set_string("file-autoload-file", w.get_filename()))
        self.config.connect("changed::autoload-file", lambda w, fname: self.button_autoload_file.set_filename(w.get_string(fname)))
        self.button_autoload_file.set_sensitive(self.check_autoload.get_active())

        filter = Gtk.FileFilter()
        filter.set_name(_('Revelation files'))
        filter.add_mime_type("application/x-revelation")
        self.button_autoload_file.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All files'))
        filter.add_pattern("*")
        self.button_autoload_file.add_filter(filter)

        eventbox = ui.EventBox(self.button_autoload_file)
        eventbox.set_tooltip_text(_('File to open when Revelation is started'))

        hbox = ui.HBox()
        hbox.pack_start(self.check_autoload, False, False, 0)
        hbox.pack_start(eventbox, True, True, 0)
        self.section_files.append_widget(None, hbox)

        # check-button for autosave
        self.check_autosave = Gtk.CheckButton(label=_('Automatically save data when changed'))
        self.config.bind("file-autosave", self.check_autosave, "active", Gio.SettingsBindFlags.DEFAULT)

        self.check_autosave.set_tooltip_text(_('Automatically save the data file when an entry is added, modified or removed'))
        self.section_files.append_widget(None, self.check_autosave)

        # autolock file
        self.check_autolock = Gtk.CheckButton(label=_('Lock file when inactive for'))
        self.config.bind("file-autolock", self.check_autolock, "active", Gio.SettingsBindFlags.DEFAULT)
        self.check_autolock.connect("toggled", lambda w: self.spin_autolock_timeout.set_sensitive(w.get_active()))
        self.check_autolock.set_tooltip_text(_('Automatically lock the data file after a period of inactivity'))

        self.spin_autolock_timeout = ui.SpinEntry()
        self.spin_autolock_timeout.set_range(1, 120)
        self.spin_autolock_timeout.set_sensitive(self.check_autolock.get_active())
        self.config.bind("file-autolock-timeout", self.spin_autolock_timeout, "value", Gio.SettingsBindFlags.DEFAULT)
        self.spin_autolock_timeout.set_tooltip_text(_('The period of inactivity before locking the file, in minutes'))

        hbox = ui.HBox()
        hbox.set_spacing(3)
        hbox.pack_start(self.check_autolock, False, False, 0)
        hbox.pack_start(self.spin_autolock_timeout, False, False, 0)
        hbox.pack_start(ui.Label(_('minutes')), True, True, 0)
        self.section_files.append_widget(None, hbox)


    def __init_section_gotocmd(self, page):
        "Sets up the goto command section"

        self.section_gotocmd = page.add_section(_('Goto Commands'))

        for entrytype in entry.ENTRYLIST:
            if entrytype == entry.FolderEntry:
                continue

            e = entrytype()

            widget = ui.Entry()
            self.config.bind("launcher-"+e.id, widget, "text", Gio.SettingsBindFlags.DEFAULT)

            tooltip = _('Goto command for %s accounts. The following expansion variables can be used:') % e.typename + "\n\n"

            for field in e.fields:
                tooltip += "%%%s: %s\n" % ( field.symbol, field.name )

            tooltip += "\n"
            tooltip += _('%%: a "%" sign') + "\n"
            tooltip += _('%?x: optional expansion variable') + "\n"
            tooltip += _('%(...%): optional substring expansion')

            widget.set_tooltip_text(tooltip)
            self.section_gotocmd.append_widget(e.typename, widget)


    def __init_section_password(self, page):
        "Sets up the password section"

        self.section_password = page.add_section(_('Passwords'))

        # show passwords checkbutton
        self.check_show_passwords = Gtk.CheckButton(label=_('Display passwords and other secrets'))
        self.config.bind("view-passwords", self.check_show_passwords, "active", Gio.SettingsBindFlags.DEFAULT)

        self.check_show_passwords.set_tooltip_text(_('Display passwords and other secrets, such as PIN codes (otherwise, hide with ******)'))
        self.section_password.append_widget(None, self.check_show_passwords)

        # chain username checkbutton
        self.check_chain_username = Gtk.CheckButton(label=_('Also copy username when copying password'))
        self.config.bind("clipboard-chain-username", self.check_chain_username, "active", Gio.SettingsBindFlags.DEFAULT)

        self.check_chain_username.set_tooltip_text(_('When the password is copied to clipboard, put the username before the password as a clipboard "chain"'))
        self.section_password.append_widget(None, self.check_chain_username)

        # use punctuation chars checkbutton
        self.check_punctuation_chars = Gtk.CheckButton(label=_('Use punctuation characters for passwords'))
        self.config.bind("passwordgen-punctuation", self.check_punctuation_chars, "active", Gio.SettingsBindFlags.DEFAULT)

        self.check_punctuation_chars.set_tooltip_text(_('When passwords are generated, use punctuation characters like %, =, { or .'))
        self.section_password.append_widget(None, self.check_punctuation_chars)

        # password length spinbutton
        self.spin_pwlen = ui.SpinEntry()
        self.spin_pwlen.set_range(4, 32)
        self.config.bind("passwordgen-length", self.spin_pwlen, "value", Gio.SettingsBindFlags.DEFAULT)

        self.spin_pwlen.set_tooltip_text(_('The number of characters in generated passwords - 8 or more are recommended'))
        self.section_password.append_widget(_('Length of generated passwords'), self.spin_pwlen)


    def __init_section_toolbar(self, page):
        "Sets up the toolbar section"

        self.section_toolbar = page.add_section(_('Toolbar Style'))

        # radio-button for desktop default
        self.radio_toolbar_desktop = Gtk.RadioButton.new_with_label_from_widget(None, _('Use desktop default'))
        self.radio_toolbar_desktop.connect("toggled", lambda w: w.get_active() and self.config.set_string("view-toolbar-style", "desktop"))

        self.radio_toolbar_desktop.set_tooltip_text(_('Show toolbar items with default style'))
        self.section_toolbar.append_widget(None, self.radio_toolbar_desktop)

        # radio-button for icons and text
        self.radio_toolbar_both = Gtk.RadioButton.new_with_label_from_widget(self.radio_toolbar_desktop, _('Show icons and text'))
        self.radio_toolbar_both.connect("toggled", lambda w: w.get_active() and self.config.set_string("view-toolbar-style", "both"))

        self.radio_toolbar_both.set_tooltip_text(_('Show toolbar items with both icons and text'))
        self.section_toolbar.append_widget(None, self.radio_toolbar_both)

        # radio-button for icons and important text
        self.radio_toolbar_bothhoriz = Gtk.RadioButton.new_with_label_from_widget(self.radio_toolbar_desktop, _('Show icons and important text'))
        self.radio_toolbar_bothhoriz.connect("toggled", lambda w: w.get_active() and self.config.set_string("view-toolbar-style", "both-horiz"))

        self.radio_toolbar_bothhoriz.set_tooltip_text(_('Show toolbar items with text beside important icons'))
        self.section_toolbar.append_widget(None, self.radio_toolbar_bothhoriz)

        # radio-button for icons only
        self.radio_toolbar_icons = Gtk.RadioButton.new_with_label_from_widget(self.radio_toolbar_desktop, _('Show icons only'))
        self.radio_toolbar_icons.connect("toggled", lambda w: w.get_active() and self.config.set_string("view-toolbar-style", "icons"))

        self.radio_toolbar_icons.set_tooltip_text(_('Show toolbar items with icons only'))
        self.section_toolbar.append_widget(None, self.radio_toolbar_icons)

        # radio-button for text only
        self.radio_toolbar_text = Gtk.RadioButton.new_with_label_from_widget(self.radio_toolbar_desktop, _('Show text only'))
        self.radio_toolbar_text.connect("toggled", lambda w: w.get_active() and self.config.set_string("view-toolbar-style", "text"))

        self.radio_toolbar_text.set_tooltip_text(_('Show toolbar items with text only'))
        self.section_toolbar.append_widget(None, self.radio_toolbar_text)

        {"desktop":    self.radio_toolbar_desktop,
         "both":       self.radio_toolbar_both,
         "both-horiz": self.radio_toolbar_bothhoriz,
         "icons":      self.radio_toolbar_icons,
         "text":       self.radio_toolbar_text
        }[self.config.get_string("view-toolbar-style")].set_active(True)


    def run(self):
        "Runs the preference dialog"

        self.show_all()

        # for some reason, Gtk crashes on close-by-escape unless we do this
        self.get_widget_for_response(Gtk.ResponseType.CLOSE).grab_focus()
        self.notebook.grab_focus()



if __name__ == "__main__":
    app = Revelation()
    app.set_flags(Gio.ApplicationFlags.NON_UNIQUE)
    app.run()




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for handling PasswordSafe data
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import data, entry

import locale, re, struct
from Cryptodome.Cipher import Blowfish
import Cryptodome.Random as Random


FIELDTYPE_NAME      = 0x00
FIELDTYPE_UUID      = 0x01
FIELDTYPE_GROUP     = 0x02
FIELDTYPE_TITLE     = 0x03
FIELDTYPE_USER      = 0x04
FIELDTYPE_NOTES     = 0x05
FIELDTYPE_PASSWORD  = 0x06
FIELDTYPE_END       = 0xff



# We need our own SHA1-implementation, because Password Safe does
# non-standard things we need to replicate. This implementation is
# written by J. Hallen and L. Creighton for the Pypy project, with
# slight modifications by Erik Grinaker.
class SHA:

    K = [
        0x5A827999,
        0x6ED9EBA1,
        0x8F1BBCDC,
        0xCA62C1D6
    ]


    def __init__(self, input = None):
        self.count = [0, 0]
        self.init()

        if input != None:
            self.update(input)


    def __bytelist2longBigEndian(self, list):
        imax = len(list)//4
        hl = [0] * imax

        j = 0
        i = 0
        while i < imax:
            b0 = list[j] << 24
            b1 = list[j+1] << 16
            b2 = list[j+2] << 8
            b3 = list[j+3]
            hl[i] = b0 | b1 | b2 | b3
            i = i+1
            j = j+4

        return hl


    def __long2bytesBigEndian(self, n, blocksize=0):
        s = b''
        pack = struct.pack
        while n > 0:
            s = pack('>I', n & 0xffffffff) + s
            n = n >> 32

        for i in range(len(s)):
            if s[i] != b'\000':
                break
        else:
            s = b'\000'
            i = 0

        s = s[i:]

        if blocksize > 0 and len(s) % blocksize:
            s = (blocksize - len(s) % blocksize) * b'\000' + s

        return s


    def __rotateLeft(self, x, n):
        return (x << n) | (x >> (32-n))


    def __transform(self, W):
        for t in range(16, 80):
            W.append(self.__rotateLeft(
                W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16], 1) & 0xffffffff)

        A = self.H0
        B = self.H1
        C = self.H2
        D = self.H3
        E = self.H4

        for t in range(0, 20):
            TEMP = self.__rotateLeft(A, 5) + ((B & C) | ((~ B) & D)) + E + W[t] + self.K[0]
            E = D
            D = C
            C = self.__rotateLeft(B, 30) & 0xffffffff
            B = A
            A = TEMP & 0xffffffff

        for t in range(20, 40):
            TEMP = self.__rotateLeft(A, 5) + (B ^ C ^ D) + E + W[t] + self.K[1]
            E = D
            D = C
            C = self.__rotateLeft(B, 30) & 0xffffffff
            B = A
            A = TEMP & 0xffffffff

        for t in range(40, 60):
            TEMP = self.__rotateLeft(A, 5) + ((B & C) | (B & D) | (C & D)) + E + W[t] + self.K[2]
            E = D
            D = C
            C = self.__rotateLeft(B, 30) & 0xffffffff
            B = A
            A = TEMP & 0xffffffff

        for t in range(60, 80):
            TEMP = self.__rotateLeft(A, 5) + (B ^ C ^ D)  + E + W[t] + self.K[3]
            E = D
            D = C
            C = self.__rotateLeft(B, 30) & 0xffffffff
            B = A
            A = TEMP & 0xffffffff


        self.H0 = (self.H0 + A) & 0xffffffff
        self.H1 = (self.H1 + B) & 0xffffffff
        self.H2 = (self.H2 + C) & 0xffffffff
        self.H3 = (self.H3 + D) & 0xffffffff
        self.H4 = (self.H4 + E) & 0xffffffff


    def digest(self):
        H0 = self.H0
        H1 = self.H1
        H2 = self.H2
        H3 = self.H3
        H4 = self.H4
        input = [] + self.input
        count = [] + self.count

        index = (self.count[1] >> 3) & 0x3f

        if index < 56:
            padLen = 56 - index
        else:
            padLen = 120 - index

        padding = [128] + [0] * 63
        self.update(padding[:padLen])
        bits = self.__bytelist2longBigEndian(self.input[:56]) + count

        self.__transform(bits)

        digest = self.__long2bytesBigEndian(self.H0, 4) + \
                self.__long2bytesBigEndian(self.H1, 4) + \
                self.__long2bytesBigEndian(self.H2, 4) + \
                self.__long2bytesBigEndian(self.H3, 4) + \
                self.__long2bytesBigEndian(self.H4, 4)

        self.H0 = H0
        self.H1 = H1
        self.H2 = H2
        self.H3 = H3
        self.H4 = H4
        self.input = input
        self.count = count

        return digest


    def hexdigest(self):
        return ''.join(['%02x' % ord(c) for c in self.digest()])


    def init(self, H0 = 0x67452301, H1 = 0xEFCDAB89, H2 = 0x98BADCFE, H3 = 0x10325476, H4 = 0xC3D2E1F0):
        self.length = 0
        self.input = []

        self.H0 = H0
        self.H1 = H1
        self.H2 = H2
        self.H3 = H3
        self.H4 = H4


    def update(self, inBuf):
        leninBuf = len(inBuf)

        index = (self.count[1] >> 3) & 0x3F

        self.count[1] = self.count[1] + (leninBuf << 3)
        if self.count[1] < (leninBuf << 3):
            self.count[0] = self.count[0] + 1
        self.count[0] = self.count[0] + (leninBuf >> 29)

        partLen = 64 - index

        if leninBuf >= partLen:
            self.input[index:] = list(inBuf[:partLen])
            self.__transform(self.__bytelist2longBigEndian(self.input))
            i = partLen
            while i + 63 < leninBuf:
                self.__transform(self.__bytelist2longBigEndian(list(inBuf[i:i+64])))
                i = i + 64
            else:
                self.input = list(inBuf[i:leninBuf])
        else:
            i = 0
            self.input = self.input + list(inBuf)


# misc functions common to both datahandlers
def decrypt(key, ciphertext, iv = None):
    "Decrypts data"

    if len(ciphertext) % 8 != 0:
        raise base.FormatError

    cipher      = Blowfish.new(key, Blowfish.MODE_ECB)
    cbc     = iv
    plaintext   = b""

    for cipherblock in [ ciphertext[i * 8 : (i + 1) * 8] for i in range(len(ciphertext) // 8) ]:

        plainblock = decrypt_block(cipher, cipherblock)

        if cbc != None:
            plainblock = bytes([ plainblock[i] ^ cbc[i] for i in range(len(plainblock)) ])
            cbc = cipherblock

        plaintext += plainblock

    return plaintext


def decrypt_block(cipher, block):
    "Decrypts a block with the given cipher"

    block = bytes((block[3], block[2], block[1], block[0], block[7], block[6], block[5], block[4]))
    block = cipher.decrypt(block)
    block = bytes((block[3], block[2], block[1], block[0], block[7], block[6], block[5], block[4]))

    return block


def encrypt(key, plaintext, iv = None):
    "Encrypts data"

    if len(plaintext) % 8 != 0:
        raise base.FormatError

    cipher      = Blowfish.new(key, Blowfish.MODE_ECB)
    cbc     = iv
    ciphertext  = b""

    for plainblock in [ plaintext[i * 8 : (i + 1) * 8] for i in range(len(plaintext) // 8) ]:

        if cbc != None:
            plainblock = bytes([ plainblock[i] ^ cbc[i] for i in range(len(plainblock)) ])

        cipherblock = encrypt_block(cipher, plainblock)
        ciphertext += cipherblock

        if cbc != None:
            cbc = cipherblock


    return ciphertext


def encrypt_block(cipher, block):
    "Encrypts a block with the given cipher"

    block = bytes((block[3], block[2], block[1], block[0], block[7], block[6], block[5], block[4]))
    block = cipher.encrypt(block)
    block = bytes((block[3], block[2], block[1], block[0], block[7], block[6], block[5], block[4]))

    return block


def generate_testhash(password, random):
    "Generates a testhash based on a password and a random string"

    key = SHA(random + b"\x00\x00" + password.encode()).digest()
    cipher  = Blowfish.new(key, Blowfish.MODE_ECB)

    for i in range(1000):
        random = encrypt_block(cipher, random)

    h = SHA()
    h.init(0, 0, 0, 0, 0)
    h.update(random)
    h.update(b"\x00\x00")
    testhash = h.digest()

    return testhash



def create_field(value, type = FIELDTYPE_NAME):
    "Creates a field"
    if isinstance(value, str):
        value = value.encode()

    field = struct.pack("ii", len(value), type) + value

    if len(value) == 0 or len(value) % 8 != 0:
        field += b"\x00" * (8 - len(value) % 8)

    return field


def normalize_field(field):
    "Normalizes a field value"

    enc = locale.getpreferredencoding()

    field = field.replace(b"\x00", b"")
    field = re.sub(b"\s+", b" ", field)
    field = field.strip()
    field = field.decode(enc, "replace")

    return field


def parse_field_header(header):
    "Parses field data, returns the length and type"

    if len(header) < 8:
        raise base.FormatError

    length, type = struct.unpack("ii", header[:8])

    if length == 0 or length % 8 != 0:
        length += 8 - length % 8

    return length, type



class PasswordSafe1(base.DataHandler):
    "Data handler for PasswordSafe 1.x data"

    name        = "Password Safe 1.x"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        base.DataHandler.__init__(self)


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        if len(input) < 56:
            raise base.FormatError

        if (len(input) - 56) % 8 != 0:
            raise base.FormatError


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # serialize data
        enc = locale.getpreferredencoding()
        db = b""
        iter = entrystore.iter_children(None)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                e = e.convert_generic()

                edata = b""
                edata += create_field(e.name.encode(enc, "replace") + b"\xAD" + e[entry.UsernameField].encode("iso-8859-1"))
                edata += create_field(e[entry.PasswordField].encode(enc, "replace"))
                edata += create_field(e.description.encode(enc, "replace"))

                db += edata

            iter = entrystore.iter_traverse_next(iter)


        # encrypt data
        rand = Random.new()
        random = rand.read(8)
        salt   = rand.read(20)
        iv     = rand.read(8)

        testhash    = generate_testhash(password, random)
        ciphertext  = encrypt(SHA(password.encode() + salt).digest(), db, iv)

        return random + testhash + salt + iv + ciphertext


    def import_data(self, input, password):
        "Imports data into an entrystore"

        # read header and test password
        if password is None:
            raise base.PasswordError

        random      = input[0:8]
        testhash    = input[8:28]
        salt        = input[28:48]
        iv      = input[48:56]

        if testhash != generate_testhash(password, random):
            raise base.PasswordError

        # load data
        db      = decrypt(SHA(password.encode() + salt).digest(), input[56:], iv)
        entrystore  = data.EntryStore()

        while len(db) > 0:

            dbentry = { "name" : "", "username" : "", "password" : "", "note" : "" }

            for f in ( "name", "password", "note" ):
                flen, ftype = parse_field_header(db[:8])
                value = db[8:8 + flen]

                if f == "name" and b"\xAD" in value:
                    value, dbentry["username"] = value.split(b"\xAD", 1)

                dbentry[f] = value
                db = db[8 + flen:]

            e = entry.GenericEntry()
            e.name          = normalize_field(dbentry["name"])
            e.description       = normalize_field(dbentry["note"])
            e[entry.UsernameField]  = normalize_field(dbentry["username"])
            e[entry.PasswordField]  = normalize_field(dbentry["password"])

            entrystore.add_entry(e)

        return entrystore



class PasswordSafe2(base.DataHandler):
    "Data handler for PasswordSafe 2.x data"

    name        = "Password Safe 2.x"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        base.DataHandler.__init__(self)


    def __get_group(self, entrystore, iter):
        "Returns the group path for an iter"

        path = []

        iter = entrystore.iter_parent(iter)

        while iter is not None:
            path.append(entrystore.get_entry(iter).name)
            iter = entrystore.iter_parent(iter)

        path.reverse()

        return ".".join(path)


    def __setup_group(self, entrystore, groupmap, group):
        "Sets up a group folder, or returns an existing one"

        if group in ( None, "" ):
            return None

        if group in groupmap:
            return groupmap[group]

        if "." in group:
            parent, groupname = group.rsplit(".", 1)
            parentiter = self.__setup_group(entrystore, groupmap, parent)

        else:
            groupname = group
            parentiter = None

        e = entry.FolderEntry()
        e.name = groupname

        iter = entrystore.add_entry(e, parentiter)
        groupmap[group] = iter

        return iter


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        if len(input) < 56:
            raise base.FormatError

        if (len(input) - 56) % 8 != 0:
            raise base.FormatError


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # set up magic entry at start of database
        db = b""
        db += b"\x48\x00\x00\x00\x00\x00\x00\x00"
        db += b" !!!Version 2 File Format!!! Please upgrade to PasswordSafe 2.0 or later"
        db += b"\x03\x00\x00\x00\x06\x00\x00\x00"
        db += b"2.0\x00\x00\x00\x00\x00"
        db += b"\x00\x00\x00\x00\x06\x00\x00\x00"
        db += b"\x00\x00\x00\x00\x00\x00\x00\x00"

        # serialize data
        uuids = []
        iter = entrystore.iter_children(None)

        enc = locale.getpreferredencoding()
        rand = Random.new()

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                e = e.convert_generic()

                uuid = rand.read(16)

                while uuid in uuids:
                    uuid = rand.read(16)

                edata = b""
                edata += create_field(uuid, FIELDTYPE_UUID)
                edata += create_field(self.__get_group(entrystore, iter), FIELDTYPE_GROUP)
                edata += create_field(e.name.encode(enc, "replace"), FIELDTYPE_TITLE)
                s = e[entry.UsernameField]
                if s is None:
                    s = ""
                edata += create_field(s.encode(enc, "replace"), FIELDTYPE_USER)
                edata += create_field(e[entry.PasswordField].encode(enc, "replace"), FIELDTYPE_PASSWORD)
                edata += create_field(e.description.encode(enc, "replace"), FIELDTYPE_NOTES)
                edata += create_field("", FIELDTYPE_END)

                db += edata

            iter = entrystore.iter_traverse_next(iter)


        # encrypt data
        rand = Random.new()
        random  = rand.read(8)
        salt    = rand.read(20)
        iv      = rand.read(8)

        testhash    = generate_testhash(password, random)
        ciphertext  = encrypt(SHA(password.encode() + salt).digest(), db, iv)

        return random + testhash + salt + iv + ciphertext


    def import_data(self, input, password):
        "Imports data into an entrystore"

        # read header and test password
        if password is None:
            raise base.PasswordError

        random      = input[0:8]
        testhash    = input[8:28]
        salt        = input[28:48]
        iv      = input[48:56]

        if testhash != generate_testhash(password, random):
            raise base.PasswordError

        # load data
        db      = decrypt(SHA(password.encode() + salt).digest(), input[56:], iv)
        entrystore  = data.EntryStore()


        # read magic entry
        for f in "magic", "version", "prefs":
            flen, ftype = parse_field_header(db)
            value = db[8:8 + flen]

            if f == "magic" and value != b" !!!Version 2 File Format!!! Please upgrade to PasswordSafe 2.0 or later":
                raise base.FormatError

            db = db[8 + flen:]

        # import entries
        e = entry.GenericEntry()
        group = None
        groupmap = {}

        while len(db) > 0:
            flen, ftype = parse_field_header(db)
            value = normalize_field(db[8:8 + flen])

            if ftype == FIELDTYPE_NAME:
                if b"\xAD" not in value:
                    e.name = value

                else:
                    n, u = value.split(b"\xAD", 1)

                    e.name = normalize_field(n)
                    e[entry.UsernameField] = normalize_field(u)

            elif ftype == FIELDTYPE_TITLE:
                e.name = value

            elif ftype == FIELDTYPE_USER:
                e[entry.UsernameField] = value

            elif ftype == FIELDTYPE_PASSWORD:
                e[entry.PasswordField] = value

            elif ftype == FIELDTYPE_NOTES:
                e.description = value

            elif ftype == FIELDTYPE_UUID:
                pass

            elif ftype == FIELDTYPE_GROUP:
                group = value

            elif ftype == FIELDTYPE_END:
                if group not in ( None, "" ):
                    parent = self.__setup_group(entrystore, groupmap, group)

                else:
                    parent = None

                entrystore.add_entry(e, parent)

                e = entry.GenericEntry()
                group = None

            else:
                pass

            db = db[8 + flen:]

        return entrystore



class MyPasswordSafe(PasswordSafe2):
    "Data handler for MyPasswordSafe data"

    name        = "MyPasswordSafe"
    importer    = True
    exporter    = True
    encryption  = True



class MyPasswordSafeOld(PasswordSafe1):
    "Data handler for old MyPasswordSafe data"

    name        = "MyPasswordSafe (old format)"
    importer    = True
    exporter    = True
    encryption  = True



class PasswordGorilla(PasswordSafe2):
    "Data handler for Password Gorilla data"

    name        = "Password Gorilla"
    importer    = True
    exporter    = True
    encryption  = True




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module for handling Revelation data
#
#
# Copyright (c) 2003-2006 Erik Grinaker
# Copyright (c) 2012 Mikel Olasagasti
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import config, data, entry, util
from revelation.bundle import luks

from Cryptodome.Protocol.KDF import PBKDF2
from Cryptodome.Hash import SHA1
from Cryptodome.Random import get_random_bytes

import os, re, struct, xml.dom.minidom, zlib
from io import BytesIO

from xml.parsers.expat import ExpatError
from Cryptodome.Cipher import AES

import hashlib


class RevelationXML(base.DataHandler):
    "Handler for Revelation XML data"

    name        = "XML"
    importer    = True
    exporter    = True
    encryption  = False


    def __init__(self):
        base.DataHandler.__init__(self)


    def __lookup_entry(self, typename):
        "Looks up an entry type based on an identifier"

        for entrytype in entry.ENTRYLIST:
            if entrytype().id == typename:
                return entrytype

        else:
            raise entry.EntryTypeError


    def __lookup_field(self, fieldname):
        "Looks up an entry field based on an identifier"

        for fieldtype in entry.FIELDLIST:
            if fieldtype.id == fieldname:
                return fieldtype

        else:
            raise entry.EntryFieldError


    def __xml_import_node(self, entrystore, node, parent = None):
        "Imports a node into an entrystore"

        try:

            # check the node
            if node.nodeType == node.TEXT_NODE:
                return

            if node.nodeType != node.ELEMENT_NODE or node.nodeName != "entry":
                raise base.FormatError

            # create an entry, iter needed for children
            e = self.__lookup_entry(node.attributes["type"].value)()
            iter = entrystore.add_entry(e, parent)

            # handle child nodes
            for child in node.childNodes:

                if child.nodeType != child.ELEMENT_NODE:
                    continue

                elif child.nodeName == "name":
                    e.name = util.dom_text(child)

                elif child.nodeName == "notes":
                    e.notes = util.dom_text(child)

                elif child.nodeName == "description":
                    e.description = util.dom_text(child)

                elif child.nodeName == "updated":
                    e.updated = int(util.dom_text(child))

                elif child.nodeName == "field":
                    e[self.__lookup_field(child.attributes["id"].nodeValue)] = util.dom_text(child)

                elif child.nodeName == "entry":
                    if type(e) != entry.FolderEntry:
                        raise base.DataError

                    self.__xml_import_node(entrystore, child, iter)

                else:
                    raise base.FormatError

            # update entry with actual data
            entrystore.update_entry(iter, e)


        except ( entry.EntryTypeError, entry.EntryFieldError ):
            raise base.DataError

        except KeyError:
            raise base.FormatError

        except ValueError:
            raise base.DataError


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        if isinstance(input, str):
            input = input.encode()

        match = re.match(b"""
            \s*                 # whitespace at beginning
            <\?xml(?:.*)\?>     # xml header
            \s*                 # whitespace after xml header
            <revelationdata     # open revelationdata tag
            [^>]+               # any non-closing character
            dataversion="(\d+)" # dataversion
            [^>]*               # any non-closing character
            >                   # close revelationdata tag
        """, input, re.VERBOSE)

        if match is None:
            raise base.FormatError

        if int(match.group(1)) != 1:
            raise base.VersionError


    def detect(self, input):
        "Checks if this handler can guarantee to handle some data"

        try:
            self.check(input)
            return True

        except ( base.FormatError, base.VersionError ):
            return False


    def export_data(self, entrystore, password = None, parent = None, level = 0):
        "Serializes data into an XML stream"

        xml = ""
        tabs = "\t" * (level + 1)

        for i in range(entrystore.iter_n_children(parent)):
            iter = entrystore.iter_nth_child(parent, i)
            e = entrystore.get_entry(iter)

            xml += "\n"
            xml += tabs + "<entry type=\"%s\">\n" % e.id
            xml += tabs + " <name>%s</name>\n" % util.escape_markup(e.name)
            xml += tabs + " <description>%s</description>\n" % util.escape_markup(e.description)
            xml += tabs + " <updated>%d</updated>\n" % e.updated
            xml += tabs + " <notes>%s</notes>\n" % util.escape_markup(e.notes)

            for field in e.fields:
                xml += tabs + " <field id=\"%s\">%s</field>\n" % ( field.id, util.escape_markup(field.value) )

            xml += RevelationXML.export_data(self, entrystore, password, iter, level + 1)
            xml += tabs + "</entry>\n"

        if level == 0:
            header = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"
            header += "<revelationdata version=\"%s\" dataversion=\"1\">\n" % config.VERSION
            footer = "</revelationdata>\n"

            xml = header + xml + footer

        return xml


    def import_data(self, input, password = None):
        "Imports data from a data stream to an entrystore"

        RevelationXML.check(self, input)

        try:
            dom = xml.dom.minidom.parseString(input.strip())

        except ExpatError:
            raise base.FormatError


        if dom.documentElement.nodeName != "revelationdata":
            raise base.FormatError

        if "dataversion" not in dom.documentElement.attributes:
            raise base.FormatError


        entrystore = data.EntryStore()

        for node in dom.documentElement.childNodes:
            self.__xml_import_node(entrystore, node)

        return entrystore



class Revelation(RevelationXML):
    "Handler for Revelation data"

    name        = "Revelation"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        RevelationXML.__init__(self)


    def __generate_header(self):
        "Generates a header"

        header = "rvl\x00"         # magic string
        header += "\x01"           # data version
        header += "\x00"           # separator
        header += "\x00\x04\x06"   # application version TODO
        header += "\x00\x00\x00"   # separator

        return header


    def __parse_header(self, header):
        "Parses a data header, returns the data version"

        if header is None:
            raise base.FormatError

        match = re.match(b"""
            ^               # start of header
            rvl\x00         # magic string
            (.)             # data version
            \x00            # separator
            (.{3})          # app version
            \x00\x00\x00    # separator
        """, header, re.VERBOSE)

        if match is None:
            raise base.FormatError

        return ord(match.group(1))


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        if len(input) < (12 + 16):
            raise base.FormatError

        dataversion = self.__parse_header(input[:12])

        if dataversion != 1:
            raise base.VersionError


    def detect(self, input):
        "Checks if the handler can guarantee to use the data"

        try:
            self.check(input)
            return True

        except ( base.FormatError, base.VersionError ):
            return False


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # check and pad password
        if password is None:
            raise base.PasswordError

        password = util.pad_right(password[:32], 32, "\0")

        # generate XML
        data = RevelationXML.export_data(self, entrystore)

        # compress data, and right-pad with the repeated ascii
        # value of the pad length
        data = zlib.compress(data.encode())

        padlen = 16 - (len(data) % 16)
        if padlen == 0:
            padlen = 16

        data += bytearray((padlen,)) * padlen

        # generate an initial vector for the CBC encryption
        iv = os.urandom(16)

        data = AES.new(password.encode("utf8"), AES.MODE_CBC, iv).encrypt(data)

        # encrypt the iv, and prepend it to the data with a header
        data = self.__generate_header().encode("utf8") + AES.new(password.encode("utf8"), AES.MODE_ECB).encrypt(iv) + data

        return data


    def import_data(self, input, password):
        "Imports data into an entrystore"

        # check and pad password
        if password is None:
            raise base.PasswordError

        password = util.pad_right(password[:32], 32, "\0")

        # check the data
        self.check(input)
        dataversion = self.__parse_header(input[:12])

        # handle only version 1 data files
        if dataversion != 1:
            raise base.VersionError


        cipher = AES.new(password.encode("utf8"), AES.MODE_ECB)
        iv = cipher.decrypt(input[12:28])


        # decrypt the data
        input = input[28:]

        if len(input) % 16 != 0:
            raise base.FormatError

        cipher = AES.new(password.encode("utf8"), AES.MODE_CBC, iv)
        input = cipher.decrypt(input)


        # decompress data
        padlen = input[-1]
        for i in input[-padlen:]:
            if i != padlen:
                raise base.PasswordError

        input = zlib.decompress(input[0:-padlen]).decode()


        # check and import data
        if input.strip()[:5] != "<?xml":
            raise base.PasswordError

        entrystore = RevelationXML.import_data(self, input)

        return entrystore


class Revelation2(RevelationXML):
    "Handler for Revelation data version 2"

    name        = "Revelation2"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        RevelationXML.__init__(self)


    def __generate_header(self):
        "Generates a header"

        header =  b"rvl\x00"        # magic string
        header += b"\x02"           # data version
        header += b"\x00"           # separator
        header += b"\x00\x04\x07"   # application version
        header += b"\x00\x00\x00"   # separator

        return header


    def __parse_header(self, header):
        "Parses a data header, returns the data version"

        if header is None:
            raise base.FormatError

        match = re.match(b"""
            ^               # start of header
            rvl\x00         # magic string
            (.)             # data version
            \x00            # separator
            (.{3})          # app version
            \x00\x00\x00    # separator
        """, header, re.VERBOSE)

        if match is None:
            raise base.FormatError

        return ord(match.group(1))


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        if len(input) < (12 + 16):
            raise base.FormatError

        dataversion = self.__parse_header(input[:12])

        if dataversion != 2:
            raise base.VersionError


    def detect(self, input):
        "Checks if the handler can guarantee to use the data"

        try:
            self.check(input)
            return True

        except ( base.FormatError, base.VersionError ):
            return False


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # check and hash password with a salt
        if password is None:
            raise base.PasswordError

        # 64-bit salt
        salt = get_random_bytes(8)

        # 256-bit key
        key = PBKDF2(password, salt, 32, count=12000, hmac_hash_module=SHA1)

        # generate XML
        data = RevelationXML.export_data(self, entrystore)

        # compress data, and right-pad with the repeated ascii
        # value of the pad length
        data = zlib.compress(data.encode())

        padlen = 16 - (len(data) % 16)
        if padlen == 0:
            padlen = 16

        data += bytearray((padlen,)) * padlen

        # 128-bit IV
        iv = get_random_bytes(16)

        data = AES.new(key, AES.MODE_CBC, iv).encrypt(hashlib.sha256(data).digest() + data)

        # encrypt the iv, and prepend it to the data with a header and the used salt
        data = self.__generate_header() + salt + iv + data

        return data

    def import_data(self, input, password):
        "Imports data into an entrystore"

        # check and pad password
        if password is None:
            raise base.PasswordError

        # check the data
        self.check(input)
        dataversion = self.__parse_header(input[:12])

        # handle only version 2 data files
        if dataversion != 2:
            raise base.VersionError

        # Fetch the used 64 bit salt
        salt = input[12:20]
        iv = input[20:36]
        key = PBKDF2(password, salt, 32, count=12000, hmac_hash_module=SHA1)
        # decrypt the data
        input = input[36:]

        if len(input) % 16 != 0:
            raise base.FormatError

        cipher = AES.new(key, AES.MODE_CBC, iv)
        input = cipher.decrypt(input)
        hash256 = input[0:32]
        data = input[32:]

        if hash256 != hashlib.sha256(data).digest():
            raise base.PasswordError

        # decompress data
        padlen = data[-1]
        for i in data[-padlen:]:
            if i != padlen:
                raise base.FormatError

        data = zlib.decompress(data[0:-padlen]).decode()

        # check and import data
        if data.strip()[:5] != "<?xml":
            raise base.FormatError

        entrystore = RevelationXML.import_data(self, data)

        return entrystore


class RevelationLUKS(RevelationXML):
    "Handler for Revelation XML using the LUKS on disk format"

    name        = "Revelation LUKS"
    importer    = True
    exporter    = True
    encryption  = True

    def __init__(self):
        RevelationXML.__init__(self)
        self.luks_header = None
        self.luks_buff = None
        self.current_slot = False


    def check(self, input):
        "Checks if the data is valid"

        if input is None:
            raise base.FormatError

        sbuf = BytesIO(input)

        l = luks.LuksFile()

        try:
            l.load_from_file(sbuf)

        except:
            l.close()
            raise base.FormatError

        l.close()


    def detect(self, input):
        "Checks if the handler can guarantee to use the data"

        try:
            self.check(input)
            return True

        except ( base.FormatError, base.VersionError ):
            return False


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        # check and pad password
        if password is None:
            raise base.PasswordError

        # generate and compress XML
        data = RevelationXML.export_data(self, entrystore)
        data = zlib.compress(data.encode())

        # data needs to be padded to 512 bytes
        # We use Merkle-Damgard length padding (1 bit followed by 0 bits + size)
        # http://en.wikipedia.org/wiki/Merkle-Damg%C3%A5rd_hash_function
        padlen = 512 - (len(data) % 512)

        if padlen < 4:
            padlen = 512 + padlen

        if padlen > 4:
            data += bytes([128] + [0] * (padlen - 5))

        data += struct.pack("<I", padlen)

        # create a new luks file in memory
        buffer      = BytesIO()
        luksfile    = luks.LuksFile()
        luksfile.create(buffer, "aes", "cbc-essiv:sha256", "sha1", 16, 400)

        luksfile.set_key(0, password, 5000, 400)

        # encrypt the data
        luksfile.encrypt_data(0, data)
        buffer.seek(0)

        return buffer.read()


    def import_data(self, input, password):
        "Imports data into an entrystore"

        # check password
        if password is None:
            raise base.PasswordError

        # create a LuksFile
        buffer      = BytesIO(input)
        luksfile    = luks.LuksFile()

        try:
            luksfile.load_from_file(buffer)

        except:
            luksfile.close()
            buffer.close()
            raise base.FormatError

        slot = luksfile.open_any_key(password)

        if slot is None:
            luksfile.close()
            buffer.close()
            raise base.PasswordError

        data = luksfile.decrypt_data(0, luksfile.data_length())

        # remove the pad, and decompress
        padlen = struct.unpack("<I", data[-4:])[0]
        data = zlib.decompress(data[0:-padlen]).decode()

        if data.strip()[:5] != "<?xml":
            raise base.FormatError

        entrystore = RevelationXML.import_data(self, data)

        return entrystore




# Gergely Nagy <greg@gnome.hu>
#
from . import base
from revelation import entry

import time
import csv
from io import StringIO


class CSV(base.DataHandler):
    "Data handler for CSV files"

    name        = "Comma Separated Values (CSV)"
    importer    = False
    exporter    = True
    encryption  = False

    def __init__(self):
        base.DataHandler.__init__(self)

    def export_data(self, entrystore, password = None):
        "Exports data to a CSV file"

        # fetch and sort entries
        entries = []
        iter = entrystore.iter_nth_child(None, 0)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                entries.append(e)

            iter = entrystore.iter_traverse_next(iter)



        stringwriter = StringIO()
        csvwriter = csv.writer(stringwriter, dialect="excel")

        keys = set()
        for e in entries:
            for f in e.fields:
                keys.add(f.name)

        # 'Email', 'Hostname', 'Password', 'URL', 'Username'
        keys = sorted(keys)

        csvwriter.writerow(['Name', 'Type', 'Description', 'Updated'] + keys)

        for e in entries:

            values = []
            for key in keys:
                value = ''
                for field in e.fields:
                    if key == field.name:
                        value = field.value
                values.append(value)

            updated = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(e.updated))
            csvwriter.writerow([e.name, e.typename, e.description, updated] + values)


        return stringwriter.getvalue()


class Bitwarden(base.DataHandler):
    "Data handler for Bitwarden Web Vault (CSV file)"

    name		= "Bitwarden Web Vault (CSV)"
    importer	= False
    exporter	= True
    encryption	= False

    bitwarden_csv_header_keys = [
        'folder',
        'favorite',
        'type',
        'name',
        'notes',
        'fields',
        'login_uri',
        'login_username',
        'login_password',
        'login_totp',
    ]

    # revelation type -> Bitwarden type
    type_mapping = {
        'Creditcard': 'note',  # Bitwarden type 'card' exists but seems unmanaged.
        'Crypto Key': 'login',
        'Database': 'login',
        'Door lock': 'note',
        'Email': 'login',
        'FTP': 'login',
        'Generic': 'login',
        'Phone': 'note',
        'Remote Desktop': 'login',
        'Shell': 'login',
        'VNC': 'login',
        'Website': 'login',
    }

    # revelation field -> Bitwarden field
    field_mapping = {
        'Name': 'name',
        'Type': 'type',
        'Description': 'notes',
        'Updated': 'notes',
        'CCV number': 'notes',
        'Card number': 'notes',
        'Card type': 'notes',
        'Certificate': 'notes',
        'Code': 'notes',
        'Database': 'notes',
        'Domain': 'notes',
        'Email': 'notes',
        'Expiry date': 'notes',
        'Hostname': 'login_uri',
        'Key File': 'notes',
        'Location': 'notes',
        'PIN': 'notes',
        'Password': 'login_password',
        'Phone number': 'notes',
        'Port number': 'notes',
        'URL': 'login_uri',
        'Username': 'login_username',
    }

    date_format = "%Y-%m-%d %H:%M:%S"

    def __init__(self):
        base.DataHandler.__init__(self)

    def export_data(self, entrystore, password = None):
        "Exports data to a Bitwarden CSV file"

        # fetch and sort entries
        entries = []
        iter = entrystore.iter_nth_child(None, 0)

        stringwriter = StringIO()
        csvwriter = csv.writer(stringwriter, dialect="excel")
        csvwriter.writerow(self.bitwarden_csv_header_keys)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                # Find the directory path of this item.
                path = ''
                parent = entrystore.iter_parent(iter)
                while parent:
                    path = entrystore.get_entry(parent).name + '/' + path
                    parent = entrystore.iter_parent(parent)

                values = []
                for key in self.bitwarden_csv_header_keys:
                    # Special fields
                    if key == 'folder':
                        values.append(path)
                        continue
                    if key == 'favorite':
                        values.append('')
                        continue
                    if key == 'name':
                        values.append(e.name)
                        continue
                    if key == 'type':
                        values.append(self.type_mapping[e.typename])
                        continue

                    value = ''

                    # Add export time as a note by default.
                    # Add e.description as a note.
                    # Add revelation updated time as a note.
                    if key == 'notes':
                        now = time.strftime(self.date_format, time.localtime())
                        updated = time.strftime(self.date_format, time.localtime(e.updated))
                        value += "Exported from Revelation on %s\n" % now
                        value += "Last Revelation Update: %s\n" % updated
                        if e.description:
                            value += "Description: %s\n" % e.description
                        if e.notes != '':
                            value += "Revelation Notes: %s\n" % e.notes

                    for field in e.fields:
                        if key == self.field_mapping[field.name]:
                            if self.field_mapping[field.name] == 'notes' and field.value:
                                value += '%s: %s\n' % (field.name, field.value)
                            else:
                                value += field.value
                    values.append(value)

                csvwriter.writerow(values)

            iter = entrystore.iter_traverse_next(iter)

        return stringwriter.getvalue()



#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id$
#
# Module containing data-related functionality
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import datahandler, entry

from gi.repository import GObject, Gtk, Gdk, GLib
import time


COLUMN_NAME  = 0
COLUMN_ICON  = 1
COLUMN_ENTRY = 2

SEARCH_NEXT     = "next"
SEARCH_PREVIOUS = "prev"



class Clipboard(GObject.GObject):
    "A normal text-clipboard"

    def __init__(self):
        GObject.GObject.__init__(self)

        self.clip_clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        self.clip_primary   = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)

        self.cleartimer     = Timer(10)
        self.cleartimeout   = 60
        self.cleartimer.connect("ring", self.__cb_clear_ring)

        self.content        = None
        self.contentpointer = 0


    def __cb_clear(self, clipboard, data = None):
        "Clears the clipboard data"

        self.clip_clipboard.clear()
        self.clip_primary.clear()


    def __cb_clear_ring(self, widget):
        "Handles cleartimer rings"

        self.content        = None
        self.contentpointer = 0
        self.set("", False)


    def __cb_get(self, clipboard, selectiondata, info, data):
        "Returns text for clipboard requests"

        if self.content is None:
            text = ""

        elif type(self.content) == list:

            if len(self.content) == 0:
                text = ""

            else:
                text = self.content[self.contentpointer]

            if self.contentpointer < len(self.content) - 1:
                self.contentpointer += 1

        else:
            text = str(self.content)

        selectiondata.set_text(text, len(text))


    def clear(self):
        "Clears the clipboard"

        self.clip_clipboard.clear()
        self.clip_primary.clear()


    def get(self):
        "Fetches text from the clipboard"

        text = self.clip_clipboard.wait_for_text()

        if text is None:
            text = ""

        return text


    def has_contents(self):
        "Checks if the clipboard has any contents"

        return self.clip_clipboard.wait_is_text_available()


    def set(self, content, secret = False):
        "Copies text to the clipboard"

        self.content        = content
        self.contentpointer = 0

        targets = [
            Gtk.TargetEntry.new( "text/plain",    0, 0 ),
            Gtk.TargetEntry.new( "STRING",        0, 0 ),
            Gtk.TargetEntry.new( "TEXT",          0, 0 ),
            Gtk.TargetEntry.new( "COMPOUND_TEXT", 0, 0 ),
            Gtk.TargetEntry.new( "UTF8_STRING",   0, 0 )
        ]

        self.clip_clipboard.set_text(' '.join(self.content), -1)
        self.clip_primary.set_text(' '.join(self.content), -1)


        if secret == True:
            self.cleartimer.start(self.cleartimeout)

        else:
            self.cleartimer.stop()



class EntryClipboard(GObject.GObject):
    "A clipboard for entries"

    def __init__(self):
        GObject.GObject.__init__(self)

        self.clipboard = Gtk.Clipboard.get_for_display(display=Gdk.Display.get_default(), selection=Gdk.Atom.intern("_REVELATION_ENTRY",False))
        self.__has_contents = False

        GLib.timeout_add(500, lambda: self.__check_contents())


    def __check_contents(self):
        "Callback which check the clipboard"

        state = self.has_contents()

        if state != self.__has_contents:
            self.emit("content-toggled", state)
            self.__has_contents = state

        return True


    def clear(self):
        "Clears the clipboard"

        self.clipboard.clear()
        self.__check_contents()


    def get(self):
        "Fetches entries from the clipboard"

        try:
            xml = self.clipboard.wait_for_text()

            if xml in ( None, "" ):
                return None

            handler = datahandler.RevelationXML()
            entrystore = handler.import_data(xml)

            return entrystore

        except datahandler.HandlerError:
            return None


    def has_contents(self):
        "Checks if the clipboard has any contents"

        return self.clipboard.wait_for_text() is not None


    def set(self, entrystore, iters):
        "Copies entries from an entrystore to the clipboard"

        copystore = EntryStore()

        for iter in entrystore.filter_parents(iters):
            copystore.import_entry(entrystore, iter)

        xml = datahandler.RevelationXML().export_data(copystore)
        self.clipboard.set_text(xml,-1)

        self.__check_contents()


GObject.signal_new("content-toggled", EntryClipboard,
                   GObject.SignalFlags.ACTION, GObject.TYPE_BOOLEAN,
                   ( GObject.TYPE_BOOLEAN, ))



class EntrySearch(GObject.GObject):
    "Handles searching in an EntryStore"

    def __init__(self, entrystore):
        GObject.GObject.__init__(self)
        self.entrystore = entrystore

        self.folders        = True
        self.namedesconly   = False
        self.casesensitive  = False


    def find(self, string, entrytype = None, offset = None, direction = SEARCH_NEXT):
        "Searches for an entry, starting at the given offset"

        iter = offset

        while True:

            if direction == SEARCH_NEXT:
                iter = self.entrystore.iter_traverse_next(iter)

            else:
                iter = self.entrystore.iter_traverse_prev(iter)

            # if we've wrapped around, return None
            if self.entrystore.get_path(iter) == self.entrystore.get_path(offset):
                return None

            if self.match(iter, string, entrytype) == True:
                return iter


    def find_all(self, string, entrytype = None):
        "Searches for all entries matching a term"

        matches = []
        iter = self.entrystore.iter_children(None)

        while iter != None:

            if self.match(iter, string, entrytype) == True:
                matches.append(iter)

            iter = self.entrystore.iter_traverse_next(iter)

        return matches


    def match(self, iter, string, entrytype = None):
        "Check if an entry matches the search criteria"

        if iter is None or not string:
            return False

        e = self.entrystore.get_entry(iter)


        # check entry type
        if type(e) == entry.FolderEntry and self.folders == False:
            return False

        if entrytype is not None and type(e) not in ( entrytype, entry.FolderEntry ):
            return False


        # check entry fields
        items = [ e.name, e.description, e.notes ]

        if self.namedesconly == False:
            items.extend([ field.value for field in e.fields if field.value != "" ])


        # run the search
        for item in items:
            if self.casesensitive == True and string in item:
                return True

            elif self.casesensitive == False and string.lower() in item.lower():
                return True

        return False



class EntryStore(Gtk.TreeStore):
    "A data structure for storing entries"

    def __init__(self):
        Gtk.TreeStore.__init__(
            self,
            GObject.TYPE_STRING,    # name
            GObject.TYPE_STRING,    # icon
            GObject.TYPE_PYOBJECT   # entry
        )

        self.changed = False
        self.connect("row-has-child-toggled", self.__cb_iter_has_child)

        self.set_sort_func(COLUMN_NAME, self.__cmp)
        self.set_sort_column_id(COLUMN_NAME, Gtk.SortType.ASCENDING)


    def __cmp(self, treemodel, iter1, iter2, user_data=None):
        name1 = treemodel.get_value(iter1, COLUMN_NAME).strip().lower()
        name2 = treemodel.get_value(iter2, COLUMN_NAME).strip().lower()

        return (name1 > name2) - (name1 < name2)


    def __cb_iter_has_child(self, widget, path, iter):
        "Callback for iters having children"

        if self.iter_n_children(iter) == 0:
            self.folder_expanded(iter, False)


    def add_entry(self, e, parent = None, sibling = None):
        "Adds an entry"

        # place after parent if it's not a folder
        if parent is not None and type(self.get_entry(parent)) != entry.FolderEntry:
            iter = self.insert_after(self.iter_parent(parent), parent)

        # place before sibling, if given
        elif sibling is not None:
            iter = self.insert_before(parent, sibling)

        # otherwise, append to parent
        else:
            iter = self.append(parent)

        self.update_entry(iter, e)
        self.changed = True

        return iter


    def clear(self):
        "Removes all entries"

        Gtk.TreeStore.clear(self)
        self.changed = False


    def copy_entry(self, iter, parent = None, sibling = None):
        "Copies an entry recursively"

        newiter = self.add_entry(self.get_entry(iter), parent, sibling)

        for i in range(self.iter_n_children(iter)):
            child = self.iter_nth_child(iter, i)
            self.copy_entry(child, newiter)

        return newiter


    def filter_parents(self, iters):
        "Removes all descendants from the list of iters"

        parents = []

        for child in iters:
            for parent in iters:
                if self.is_ancestor(parent, child):
                    break

            else:
                parents.append(child)

        return parents


    def folder_expanded(self, iter, expanded):
        "Sets the expanded state of an entry"

        e = self.get_entry(iter)

        if e is None or type(e) != entry.FolderEntry:
            return

        elif expanded == True:
            self.set_value(iter, COLUMN_ICON, e.openicon)

        else:
            self.set_value(iter, COLUMN_ICON, e.icon)


    def get_entry(self, iter):
        "Fetches data for an entry"

        if iter is None:
            return None

        e = self.get_value(iter, COLUMN_ENTRY)

        if e is None:
            return None

        else:
            return e.copy()


    def get_iter(self, path):
        "Gets an iter from a path"

        try:
            if not path:
                return None

            if isinstance(path, list):
                path = tuple(path)

            return Gtk.TreeStore.get_iter(self, path)

        except ValueError:
            return None


    def get_path(self, iter):
        "Gets a path from an iter"

        return iter is not None and Gtk.TreeStore.get_path(self, iter) or None


    def get_popular_values(self, fieldtype, threshold = 3):
        "Gets popular values for a field type"

        valuecount = {}
        iter = self.iter_nth_child(None, 0)

        while iter is not None:
            e = self.get_entry(iter)

            if e.has_field(fieldtype) == False:
                iter = self.iter_traverse_next(iter)
                continue

            value = e[fieldtype].strip()

            if value != "":
                if value not in valuecount:
                    valuecount[value] = 0

                valuecount[value] += 1

            iter = self.iter_traverse_next(iter)

        popular = [ value for value, count in valuecount.items() if count >= threshold ]
        popular.sort()

        return popular


    def import_entry(self, source, iter, parent = None, sibling = None):
        "Recursively copies an entry from a different entrystore"

        if iter is not None:
            copy = self.add_entry(source.get_entry(iter), parent, sibling)
            parent, sibling = copy, None

        else:
            copy = None

        newiters = []
        for i in range(source.iter_n_children(iter)):
            child = source.iter_nth_child(iter, i)
            newiter = self.import_entry(source, child, parent, sibling)
            newiters.append(newiter)

        return copy is not None and copy or newiters


    def iter_traverse_next(self, iter):
        "Gets the 'logically next' iter"

        # get the first child, if any
        child = self.iter_nth_child(iter, 0)
        if child is not None:
            return child

        # check for a sibling or, if not found, a sibling of any ancestors
        parent = iter
        while parent is not None:
            sibling = parent.copy()
            sibling = self.iter_next(sibling)

            if sibling is not None:
                return sibling

            parent = self.iter_parent(parent)

        return None


    def iter_traverse_prev(self, iter):
        "Gets the 'logically previous' iter"

        # get the previous sibling, or parent, of the iter - if any
        if iter is not None:
            parent = self.iter_parent(iter)
            index = self.get_path(iter)[-1]

            # if no sibling is found, return the parent
            if index == 0:
                return parent

            # otherwise, get the sibling
            iter = self.iter_nth_child(parent, index - 1)

        # get the last, deepest child of the sibling or root, if any
        while self.iter_n_children(iter) > 0:
            iter = self.iter_nth_child(iter, self.iter_n_children(iter) - 1)

        return iter


    def move_entry(self, iter, parent = None, sibling = None):
        "Moves an entry"

        newiter = self.copy_entry(iter, parent, sibling)
        self.remove_entry(iter)

        return newiter


    def remove_entry(self, iter):
        "Removes an entry, and its children if any"

        if iter is None:
            return None

        self.remove(iter)
        self.changed = True


    def update_entry(self, iter, e):
        "Updates an entry"

        if None in ( iter, e):
            return None

        self.set_value(iter, COLUMN_NAME, e.name)
        self.set_value(iter, COLUMN_ICON, e.icon)
        self.set_value(iter, COLUMN_ENTRY, e.copy())

        self.changed = True



class Timer(GObject.GObject):
    "Handles timeouts etc"

    def __init__(self, resolution = 1):
        GObject.GObject.__init__(self)

        self.offset  = None
        self.timeout = None

        GLib.timeout_add(resolution * 1000, self.__cb_check)


    def __cb_check(self):
        "Checks if the timeout has been reached"

        if None not in (self.offset, self.timeout) and int(time.time()) >= (self.offset + self.timeout):
            self.stop()
            self.emit("ring")

        return True


    def reset(self):
        "Resets the timer"

        if self.offset != None:
            self.offset = int(time.time())


    def start(self, timeout):
        "Starts the timer"

        if timeout == 0:
            self.stop()

        else:
            self.offset = int(time.time())
            self.timeout = timeout


    def stop(self):
        "Stops the timer"

        self.offset = None
        self.timeout = None


GObject.signal_new("ring", Timer, GObject.SignalFlags.ACTION,
                   GObject.TYPE_BOOLEAN, ())



class UndoQueue(GObject.GObject):
    "Handles undo/redo tracking"

    def __init__(self):
        GObject.GObject.__init__(self)

        self.queue  = []
        self.pointer    = 0


    def add_action(self, name, cb_undo, cb_redo, actiondata):
        "Adds an action to the undo queue"

        del self.queue[self.pointer:]

        self.queue.append(( name, cb_undo, cb_redo, actiondata ))
        self.pointer = len(self.queue)

        self.emit("changed")


    def can_redo(self):
        "Checks if a redo action is possible"

        return self.pointer < len(self.queue)


    def can_undo(self):
        "Checks if an undo action is possible"

        return self.pointer > 0


    def clear(self):
        "Clears the queue"

        self.queue = []
        self.pointer = 0

        self.emit("changed")


    def get_redo_action(self):
        "Returns data for the next redo operation"

        if self.can_redo() == False:
            return None

        name, cb_undo, cb_redo, actiondata = self.queue[self.pointer]

        return cb_redo, name, actiondata


    def get_undo_action(self):
        "Returns data for the next undo operation"

        if self.can_undo() == False:
            return None

        name, cb_undo, cb_redo, actiondata = self.queue[self.pointer - 1]

        return cb_undo, name, actiondata


    def redo(self):
        "Executes a redo operation"

        if self.can_redo() == False:
            return None

        cb_redo, name, actiondata = self.get_redo_action()
        self.pointer += 1

        cb_redo(name, actiondata)
        self.emit("changed")


    def undo(self):
        "Executes an undo operation"

        if self.can_undo() == False:
            return None

        cb_undo, name, actiondata = self.get_undo_action()
        self.pointer -= 1

        cb_undo(name, actiondata)
        self.emit("changed")


GObject.type_register(UndoQueue)
GObject.signal_new("changed", UndoQueue, GObject.SignalFlags.ACTION,
                   GObject.TYPE_BOOLEAN, ())




#
# Revelation - a password manager for GNOME 2
# http://oss.codepoet.no/revelation/
# $Id: gpass.py 602 2007-01-03 08:06:28Z erikg $
#
# Module for handling GPass data
#
#
# Copyright (c) 2003-2006 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from . import base
from revelation import data, entry

import locale, re
from Cryptodome.Cipher import Blowfish
from Cryptodome.Hash import SHA


IV  = b"\x05\x17\x01\x7b\x0c\x03\x36\x5e"


def decrypt(ciphertext, password, magic = None):
    "Decrypts a data stream"

    # decrypt data
    if len(ciphertext) % 8 != 0:
        raise base.FormatError

    key     = SHA.new(password.encode()).digest()
    cipher      = Blowfish.new(key, Blowfish.MODE_CBC, IV)

    plaintext   = cipher.decrypt(ciphertext)

    # check magic string
    if magic != None:
        if plaintext[:len(magic)].decode() != magic:
            raise base.PasswordError

        else:
            plaintext = plaintext[len(magic):]

    # remove padding
    padchar = plaintext[-1]
    npadchar = padchar

    if (npadchar > 0):
        if plaintext[-npadchar:] != bytes([padchar] * npadchar):
            raise base.FormatError

        plaintext = plaintext[:-npadchar]

    return plaintext.decode()


def encrypt(plaintext, password):
    "Encrypts a data stream"

    # right-pad data
    padlen = 8 - len(plaintext) % 8

    if padlen == 0:
        padlen = 8

    plaintext += bytes([padlen] * padlen)

    # encrypt data
    key = SHA.new(password).digest()
    cipher  = Blowfish.new(key, Blowfish.MODE_CBC, IV)

    return cipher.encrypt(plaintext)



class GPass04(base.DataHandler):
    "Data handler for GPass 0.4.x data"

    name        = "GPass 0.4.x"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        base.DataHandler.__init__(self)


    def export_data(self, entrystore, password):
        "Exports data to a data stream"

        data = "GNOME Password Manager\n"

        iter = entrystore.iter_nth_child(None, 0)

        while iter is not None:
            e = entrystore.get_entry(iter)

            if type(e) != entry.FolderEntry:
                e = e.convert_generic()

                data += e.name + "\n"
                data += (e[entry.UsernameField] or "") + "\n"
                data += (e[entry.PasswordField] or "") + "\n"
                data += (e[entry.HostnameField] or "") + "\n"
                data += str(e.updated) + "\n"
                data += str(e.updated) + "\n"
                data += "0\n"
                data += str(len(e.description) + 1) + "\n"
                data += e.description + "\n"

            iter = entrystore.iter_traverse_next(iter)

        return encrypt(data.encode(), password.encode())


    def import_data(self, input, password):
        "Imports data from a data stream to an entrystore"

        plaintext = decrypt(input, password, "GNOME Password Manager\n")

        entrystore = data.EntryStore()
        lines = plaintext.splitlines()

        while len(lines) > 0:

            e = entry.GenericEntry()

            e.name          = lines[0]
            e[entry.UsernameField]  = lines[1]
            e[entry.PasswordField]  = lines[2]
            e[entry.HostnameField]  = lines[3]
            e.updated       = int(lines[5])
            desclen         = int(lines[7])

            if e[entry.HostnameField] == "http://":
                e[entry.HostnameField] = ""

            del lines[:8]

            d = ""
            while len(d) < desclen and len(lines) > 0:
                d += lines[0] + "\n"
                del lines[0]

            e.description = re.sub("[\r\n]+", " ", d).strip()

            entrystore.add_entry(e)

        return entrystore



class GPass05(base.DataHandler):
    "Data handler for GPass 0.5.x data"

    name        = "GPass 0.5.x (or newer)"
    importer    = True
    exporter    = True
    encryption  = True


    def __init__(self):
        base.DataHandler.__init__(self)


    def __getint(self, input):
        "Fetches an integer from the input"

        if len(input) < 4:
            raise base.FormatError

        return ord(input[0]) << 0 | ord(input[1]) << 8 | ord(input[2]) << 16 | ord(input[3]) << 24


    def __getstr(self, input):
        "Fetches a string from the input"

        length = self.__getint(input[:4])

        if len(input) < (4 + length):
            raise base.FormatError

        string = input[4:4 + length]

        if len(string) != length:
            raise base.FormatError

        return string


    def __mkint(self, input):
        "Creates a string-representation of an integer"

        string = ""

        for i in range(4):
            string += chr(input >> i * 8 & 0xff)

        return string


    def __mkstr(self, input):
        "Makes a string suitable for inclusion in the data stream"

        return self.__mkint(len(input)) + input


    def __normstr(self, string):
        "Normalizes a string"

        string = re.sub("[\r\n]+", " ", string)

        return string


    def __packint(self, input):
        "Packs an integer"

        if input == 0:
            return "\x00"

        string = ""

        while input > 0:
            c   = input % 0x80
            input   = input // 0x80

            if input > 0:
                c |= 0x80

            string += chr(c)

        return string


    def __packstr(self, input):
        "Packs a string"

        return self.__packint(len(input)) + input


    def __unpackint(self, input):
        "Fetches a packed number from the input"

        value   = 0
        b   = 1

        for i in range(min(len(input), 6)):
            c = ord(input[i])

            if c & 0x80:
                value   += b * (c & 0x7f)
                b   *= 0x80;

            else:
                value   += b * c

                return i + 1, value

        # if we didn't return in the for-loop, the input is invalid
        else:
            raise base.FormatError


    def __unpackstr(self, input):
        "Unpacks a string from the input"

        cut, length = self.__unpackint(input[:6])

        if len(input) < cut + length:
            raise base.FormatError

        return cut + length, input[cut:cut + length]


    def export_data(self, entrystore, password):
        "Exports data from an entrystore"

        plaintext   = "GPassFile version 1.1.0"
        iter        = entrystore.iter_children(None)
        id      = 0
        foldermap   = {}

        while iter != None:
            id += 1

            path        = entrystore.get_path(iter).to_string()
            parentpath  = ':'.join(path.split(':')[:-1])

            if len(parentpath) > 0 and parentpath in foldermap:
                parentid = foldermap[parentpath]

            else:
                parentid = 0


            e = entrystore.get_entry(iter)

            if type(e) == entry.FolderEntry:
                foldermap[path] = id

            elif type(e) != entry.GenericEntry:
                e = e.convert_generic()


            entrydata   = ""
            entrydata   += self.__mkint(id)
            entrydata   += self.__mkint(parentid)
            entrydata   += self.__mkstr(type(e) == entry.FolderEntry and "folder" or "general")

            attrdata    = ""
            attrdata    += self.__packstr(e.name)
            attrdata    += self.__packstr(e.description)
            attrdata    += self.__packint(e.updated)
            attrdata    += self.__packint(e.updated)
            attrdata    += self.__packint(0)
            attrdata    += self.__packint(0)

            if type(e) == entry.GenericEntry:
                attrdata    += self.__packstr(e[entry.UsernameField] or "")
                attrdata    += self.__packstr(e[entry.PasswordField] or "")
                attrdata    += self.__packstr(e[entry.HostnameField] or "")

            entrydata   += self.__mkstr(attrdata)
            plaintext   += entrydata

            iter = entrystore.iter_traverse_next(iter)

        return encrypt(plaintext.encode(), password.encode())


    def import_data(self, input, password):
        "Imports data from a data stream to an entrystore"

        plaintext = decrypt(input, password, "GPassFile version 1.1.0")

        entrystore = data.EntryStore()
        foldermap = {}

        while len(plaintext) > 0:

            # parse data
            id      = self.__getint(plaintext[:4])
            plaintext   = plaintext[4:]

            parentid    = self.__getint(plaintext[:4])
            plaintext   = plaintext[4:]

            entrytype   = self.__getstr(plaintext)
            plaintext   = plaintext[4 + len(entrytype):]

            attrdata    = self.__getstr(plaintext)
            plaintext   = plaintext[4 + len(attrdata):]


            l, name     = self.__unpackstr(attrdata)
            attrdata    = attrdata[l:]

            l, desc     = self.__unpackstr(attrdata)
            attrdata    = attrdata[l:]

            l, ctime    = self.__unpackint(attrdata)
            attrdata    = attrdata[l:]

            l, mtime    = self.__unpackint(attrdata)
            attrdata    = attrdata[l:]

            l, expire   = self.__unpackint(attrdata)
            attrdata    = attrdata[l:]

            l, etime    = self.__unpackint(attrdata)
            attrdata    = attrdata[l:]

            if entrytype == "general":
                l, username = self.__unpackstr(attrdata)
                attrdata    = attrdata[l:]

                l, password = self.__unpackstr(attrdata)
                attrdata    = attrdata[l:]

                l, hostname = self.__unpackstr(attrdata)
                attrdata    = attrdata[l:]

            else:
                username = password = hostname = ""


            # create entry
            if entrytype == "general":
                e = entry.GenericEntry()

                e.name          = self.__normstr(name)
                e.description       = self.__normstr(desc)
                e.updated       = mtime

                e[entry.HostnameField]  = self.__normstr(hostname)
                e[entry.UsernameField]  = self.__normstr(username)
                e[entry.PasswordField]  = self.__normstr(password)

            elif entrytype == "folder":
                e = entry.FolderEntry()

                e.name          = self.__normstr(name)
                e.description       = self.__normstr(desc)
                e.updated       = mtime

            else:
                continue


            # add entry to entrystore
            if parentid in foldermap:
                parent = foldermap[parentid]

            else:
                parent = None

            iter = entrystore.add_entry(e, parent)

            if type(e) == entry.FolderEntry:
                foldermap[id] = iter


        return entrystore




"""Implementation of anti-forensic information splitting

The AFsplitter supports secure data destruction crucial for secure on-disk key
management. The key idea is to bloat information and therefor improving the
chance of destroying a single bit of it. The information is bloated in such a
way, that a single missing bit causes the original information become
unrecoverable. The theory behind AFsplitter is presented in TKS1.

The interface is simple. It consists of two functions:

AFSplit(data, stripes, digesttype='sha1')
AFMerge(data, stripes, digesttype='sha1')

AFSplit operates on data and returns information splitted data.  AFMerge does
just the opposite: uses the information stored in data to recover the original
splitted data.

http://clemens.endorphin.org/AFsplitter
TKS1 paper from http://clemens.endorphin.org/publications

Copyright 2006 John Lenz <lenz@cs.wisc.edu>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/copyleft/gpl.html
"""


import hashlib, string, math, struct

import Cryptodome.Random as Random

def _xor(a, b):
    """Internal function to performs XOR on two strings a and b"""

    return bytes([x ^ y for (x, y) in zip(a, b)])

def _diffuse(block, size, digest):
    """Internal function to diffuse information inside a buffer"""

    # Compute the number of full blocks, and the size of the leftover block
    digest_size = hashlib.new(digest).digest_size
    full_blocks = int(math.floor(float(len(block)) / float(digest_size)))
    padding = len(block) % digest_size

    # hash the full blocks
    ret = b""
    for i in range(0, full_blocks):
        hash = hashlib.new(digest)
        hash.update(struct.pack(">I", i))
        hash.update(block[i*digest_size:(i+1)*digest_size])
        ret += hash.digest()

    # Hash the remaining data
    if padding > 0:
        hash = hashlib.new(digest)
        hash.update(struct.pack(">I", full_blocks))
        hash.update(block[full_blocks * digest_size:])
        ret += hash.digest()[:padding]

    return ret

def AFSplit(data, stripes, digesttype='sha1'):
    """AF-Split data using digesttype.  Returned data size will be len(data) * stripes"""

    blockSize = len(data)

    rand = Random.new()

    bufblock = [0] * blockSize

    ret = b""
    for i in range(0, stripes-1):

        # Get some random data
        r = rand.read(blockSize)

        ret += r
        bufblock = _xor(r, bufblock)
        bufblock = _diffuse(bufblock, blockSize, digesttype)

    ret += _xor(bufblock, data)
    return ret

def AFMerge(data, stripes, digesttype='sha1'):
    """AF-Merge data using digesttype.  len(data) must be a multiple of stripes"""

    if len(data) % stripes != 0:
        raise ValueError("ERROR: data is not a multiple of strips")

    blockSize = len(data) // stripes

    bufblock = [0] * blockSize
    for i in range(0, stripes - 1):
        bufblock = _xor(data[i*blockSize:(i+1)*blockSize], bufblock)
        bufblock = _diffuse(bufblock, blockSize, digesttype)

    return _xor(data[(stripes-1)*blockSize:], bufblock)



#!/usr/bin/env python3

import os
import subprocess

build_root = os.environ.get('MESON_BUILD_ROOT')
source_root = os.environ.get('MESON_SOURCE_ROOT')

print('Install schemas in build dir…')

source_datadir = os.path.join(source_root, 'data')
source_file = os.path.join(source_datadir, 'gschemas.compiled')
targetdir = os.path.join(build_root, 'data', 'glib-2.0', 'schemas')

subprocess.call(['glib-compile-schemas', source_datadir])
subprocess.call(['mkdir', '-p', targetdir])
subprocess.call(['mv', source, targetdir])
